/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : extractCode.c                                            */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Extracts the SQLCODE value and stores it in the message  */
/*                   at the location specified by the CopyValueTo attribute.  */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: July 2001                      */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "getSQLCode.h"      /* Specific header for this dll                      */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */

/* local function prototypes */
void extractSQLCode(IN CciChar *sqlErrorMessage,
					OUT int *sqlCodeP);
bool_t getSQLError(IN void *exceptionList,
				   OUT CciChar **errorMessageP,
				   OUT int *sqlCodeP);
bool_t searchTreeForSQLCode(IN CciElement *elementP,
						    OUT CciChar **errorMessageP,
						    OUT int *sqlCodeP);
int copySQLCodeToElement(IN CciElement *elementP, 
					     IN int sqlCode,
					     IN bool_t newElement,
					     IN bool_t xmlMessage);
bool_t lookForDatabaseException(IN CciElement *elementP,
						        OUT CciChar **errorMessageP,
						        OUT int *sqlCodeP);

/******************************************************************************/
/* evaluateGetSQLCode                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Extracts the SQLCODE value and stores it in the current message at the     */
/* location specified by the CopyValueTo attribute.                           */
/*                                                                            */
/******************************************************************************/
int evaluateGetSQLCode(context_t* contextP,
 			   		   CciMessage *destinationList,
		               CciMessage *exceptionList,
		               CciMessage *message)
{
	int rc,
		retCode,
		messageNumber,
		sqlCode;
	CciMessageContext *msgContextP;
	CciMessage *newMsg;
	CciElement *copyToElement;
	char mbString[MAX_STRING_SIZE + 1],
		 msgBuffer[1000],
		 nodeNameMB[100];
	deployedGetSQLCode_t *deployedGetSQLCodeP;
	bool_t newElementCreated,
		   xmlMessage,
		   sqlCodeFound;
	CciChar elementPath[10 * MAX_STRING_SIZE + 1],
		    *elementPathP,
			*dotCharP,
			*sqlErrorMessage = NULL;
    static char *functionName="evaluateGetSQLCode";

    TRACE_ENTRY("evaluateGetSQLCode", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	deployedGetSQLCodeP = (deployedGetSQLCode_t *)(contextP->nodeDataP);

	sqlCodeFound = getSQLError(exceptionList,
				               &sqlErrorMessage,
				               &sqlCode);

	if (!sqlCodeFound)
	{
		// propagate to not found terminal
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'notFound' "
			                   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_NOT_FOUND,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'notFound' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/* we do not use the extracted message, so just free it */
	if (sqlErrorMessage)
	{
		TRACE_MID(TRACE_INFO, ("Free'ing sqlErrorMessage at %.8x", sqlErrorMessage));
		free(sqlErrorMessage);
	}

	/* create a copy of message so we can insert the uuid value */
	msgContextP = cniGetMessageContext(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get message context "
			                    "(cniGetMessageContext rc=%d)", rc));
		TRACE_RETURN(RSN_014, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

    newMsg = cniCreateMessage(&rc, msgContextP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new message "
			                    "(cniCreateMessage rc=%d)", rc));
		TRACE_RETURN(RSN_015, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

	rc = copyMessage(newMsg, message);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new message"));
		TRACE_RETURN(RSN_034, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}

	elementPathP = deployedGetSQLCodeP->copyValueToLocation;
	xmlMessage = messageBodyIsXML(newMsg);
	if (xmlMessage)
	{
		/* 
		 * If path does not specify a known child element of root, then modify it if it 
		 * doesn't begin with Root.XML.
		 */
		if (UCncmp(C.ROOT_DOT_MQMD_DOT, elementPathP, UClen(C.ROOT_DOT_MQMD_DOT)) &&
		    UCncmp(C.ROOT_DOT_PROPERTIES_DOT, elementPathP, UClen(C.ROOT_DOT_PROPERTIES_DOT)))
		{
			TRACE_MID(TRACE_INFO, ("Checking path '%S' for XML qualifier", elementPathP));
			/* modify specified path if it doesn't begin with Root.XML. */
			if (UCncmp(C.ROOT_DOT_XML_DOT, elementPathP, UClen(C.ROOT_DOT_XML_DOT)))
			{
				/* specified path doesn't include XML qualifier, so modify it */
				TRACE_MID(TRACE_INFO, ("Adding XML qualifier to path"));

				/* build modified path string */
				UCcpy(elementPath, C.ROOT_DOT_XML_DOT);

				dotCharP = UCchr(elementPathP, C.DOT_CHAR);
				dotCharP++;  // move to next character following dot


				UCcat(elementPath, dotCharP);
				TRACE_MID(TRACE_INFO, ("Modified path is '%S'", elementPath));

				elementPathP = elementPath;
			}
		}
	}

	/* locate (and create if necessary) specified output element */
	copyToElement = locateElementInMessage(newMsg,
		                                   elementPathP,
									       TRUE,  /* create new element(s) if required */  
									       NULL,
									       &newElementCreated);
	/* 
	 * Propagate message to the 'failure' terminal if we can't find/create the element
	 */
    if (!copyToElement)
	{

		UCsprintf(msgBuffer, "Failed to locate specified syntax element '%S' in message",
							  deployedGetSQLCodeP->copyValueToLocation);

        cciLog(&rc, 
		       CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_FAILED_TO_LOCATE_SYNTAX_ELEMENT, 
			   msgBuffer,
#if !defined(WIN32)
		       brokerName,
#endif
			   ATTR_COPY_VALUE_TO_MB,
 			   /* convert copyToLocation to multi byte string */
			   UCToStr(deployedGetSQLCodeP->copyValueToLocation,
					   mbString,
					   sizeof(mbString),
					   NULL),
			   (char *)0);

		/* delete the new message */
		cniDeleteMessage(&rc,
						 newMsg);
		
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
			                   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_FAILURE,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/* copy the sqlCode to the located (or created) element */
	retCode = copySQLCodeToElement(copyToElement, 
		                           sqlCode, 
								   newElementCreated,
								   xmlMessage);
	if (retCode != TT_SUCCESS)
	{
		/* handle a variety of errors here... */
		if (retCode == TT_FAILED_TO_SET_ELEMENT_VALUE)
		{
			UCsprintf(msgBuffer, "Failed to store SQLCODE value in message at syntax element '%S'",
								deployedGetSQLCodeP->copyValueToLocation);
			messageNumber = MJB_FAILED_TO_STORE_SQLCODE_VALUE_IN_SYNTAX_ELEMENT;
		}
		else
		if (retCode == TT_INVALID_ELEMENT_TYPE_SPECIFIED)
		{
			UCsprintf(msgBuffer, "The specified element for storing the SQLCODE value is not "
				                 "suitable for this operation.  The syntax element is '%S'",
								deployedGetSQLCodeP->copyValueToLocation);
			messageNumber = MJB_INCOMPATIBLE_SYNTAX_ELEMENT_SPECIFIED_FOR_SQLCODE;
		}
		else
		{
			UCsprintf(msgBuffer, "Unknown error occurred when storing the SQLCODE value at the "
				                 "specified syntax element '%S'",
								deployedGetSQLCodeP->copyValueToLocation);
			messageNumber = MJB_UNKNOWN_ERROR_WHEN_STORING_SQLCODE;
		}

		UCsprintf(nodeNameMB, "'%S'", contextP->nodeName);

		cciLog(&rc, 
			   CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   messageNumber, 
			   msgBuffer,
#if !defined(WIN32)
			   brokerName,
#endif
			   ATTR_COPY_VALUE_TO_MB,
			   nodeNameMB,
 			   /* convert copyToLocation to multi byte string */
			   UCToStr(deployedGetSQLCodeP->copyValueToLocation,
					   mbString,
					   sizeof(mbString),
					   NULL),
			   (char *)0);

		/* delete the new message */
		cniDeleteMessage(&rc,
						 newMsg);
		
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
							   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_FAILURE,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/* propagate message */
	TRACE_MID(TRACE_INFO, ("Propagating message to 'out' terminal"));
	retCode = propagateMessage(contextP,
							   C.TERM_OUT,
							   newMsg,
							   destinationList,
							   exceptionList,
						       TRUE); /* this is a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		RETHROW_IF_EXCEPTION(retCode);
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	/* delete the new message */
    cniDeleteMessage(&rc,
		             newMsg);

	TRACE_RETURN(RSN_019, TT_SUCCESS);
}


/******************************************************************************/
/* getSQLError                                                                */
/******************************************************************************/
bool_t getSQLError(IN void *exceptionList,
				   OUT CciChar **errorMessageP,
				   OUT int *sqlCodeP)
{
	int rc;
	CciElement *rootElementP,
		       *currentElementP,
		       *messageElementP = NULL;
 	bool_t sqlCodeFound;

    TRACE_ENTRY("getSQLError", ("exceptionList=%.8x", exceptionList));

	/* The general structure of the exception list is shown below
	 
(
  (0x1000000)RecoverableException = (
    (0x3000000)File                 = 'xxx.cpp'
    (0x3000000)Line                 = 538
    (0x3000000)Function             = 'ImbDataFlowNode::createExceptionList'
    (0x3000000)Type                 = 'ComIbmDatabaseNode'
    (0x3000000)Name                 = '9d7160ab-e700-0000-0080-e1c5e49c6251'
    (0x3000000)Label                = 'MikesGetSQLCodeTest1.Database1'
    (0x3000000)Text                 = 'Node throwing exception'
    (0x3000000)Catalog              = 'MQSIv202'
    (0x3000000)Severity             = 3
    (0x3000000)Number               = 2230
    (0x1000000)RecoverableException = (
      (0x3000000)File              = 'xxx.cpp'
      (0x3000000)Line              = 261
      (0x3000000)Function          = 'SqlExternalDbStmt::executeStmt'
      (0x3000000)Type              = 'ComIbmDatabaseNode'
      (0x3000000)Name              = '9d7160ab-e700-0000-0080-e1c5e49c6251'
      (0x3000000)Label             = 'MikesGetSQLCodeTest1.Database1'
      (0x3000000)Text              = 'xxxx yyyy'
      (0x3000000)Catalog           = 'MQSIv202'
      (0x3000000)Severity          = 3
      (0x3000000)Number            = 2519
      (0x1000000)Insert            = (
        (0x3000000)Type = 2
        (0x3000000)Text = '1'
      )
	  ...insert list...
      (0x1000000)DatabaseException = (
        (0x3000000)File              = 'xxx.cpp'
        (0x3000000)Line              = 115
        (0x3000000)Function          = 'ImbOdbcHandle::checkRcInner'
        (0x3000000)Type              = ''
        (0x3000000)Name              = ''
        (0x3000000)Label             = ''
        (0x3000000)Text              = 'Root SQL exception'
        (0x3000000)Catalog           = 'MQSIv202'
        (0x3000000)Severity          = 3
        (0x3000000)Number            = 2321
        (0x1000000)Insert            = (
          (0x3000000)Type = 2
          (0x3000000)Text = '-1'
        )
		...insert list...
        (0x1000000)DatabaseException = (
          (0x3000000)File     = 'xxx.cpp'
          (0x3000000)Line     = 194
          (0x3000000)Function = 'ImbOdbcHandle::checkRcInner'
          (0x3000000)Type     = ''
          (0x3000000)Name     = ''
          (0x3000000)Label    = ''
          (0x3000000)Text     = 'Child SQL exception'
          (0x3000000)Catalog  = 'MQSIv202'
          (0x3000000)Severity = 3
          (0x3000000)Number   = 2322
          (0x1000000)Insert   = (
            (0x3000000)Type = 5
            (0x3000000)Text = '42S22'
          )
		  ...insert list...
        )
      )
    )
  )
)

For each DatabaseException, we scan through the insert list looking for an Insert element
with a child Text element containing the keyword SQLSTATE.  See example insert below.
(0x1000000)Insert   = (
  (0x3000000)Type = 5
  (0x3000000)Text = '[IBM][CLI Driver][DB2/NT] SQL0206N  "MESSAGE_ID" is not a column in 
  an inserted table, updated table, or any table identified in a FROM clause or is not 
  a valid transition variable for the subject table of a trigger.  SQLSTATE=42703'
)

*/
	/* Get the root element */
    rootElementP = cniRootElement(&rc, exceptionList);

	if (!rootElementP)
	{
		TRACE_MID(TRACE_INFO, ("Exception list is empty"));
		TRACE_RETURN_TYPE(RSN_001, bool_t, FALSE);
	}

	/* get first child element of Root */
	currentElementP = cniFirstChild(&rc,
								    rootElementP);

	sqlCodeFound = searchTreeForSQLCode(currentElementP,
			                            errorMessageP,
							            sqlCodeP);
	TRACE_RETURN_TYPE(RSN_001, bool_t, sqlCodeFound);
}


#define DEFAULT_TEXT_BUFFER_SIZE 10000
/******************************************************************************/
/* searchTreeForSQLCode                                                       */
/******************************************************************************/
bool_t searchTreeForSQLCode(IN CciElement *elementP,
						    OUT CciChar **errorMessageP,
						    OUT int *sqlCodeP)
{
	int rc;
	CciElement *currentElementP,
		       *messageElementP = NULL,
			   *childElementP;
    int localBufferSize = 0;
	bool_t elementHasAName = TRUE,
		   databaseExceptionElement = FALSE;

    TRACE_ENTRY("searchTreeForSQLCode", ("elementP=%.8x", elementP));

	if (lookForDatabaseException(elementP,
			                     errorMessageP,
								 sqlCodeP))
	{
		TRACE_MID(TRACE_INFO, ("DatabaseException element found"));
        TRACE_RETURN_TYPE(RSN_003, bool_t, TRUE);
	}

	/* 
	 * If we get to here, then the current element did not contain SQLCODE, so we
	 * now search its children.
	 */
    childElementP = cniFirstChild(&rc, elementP);
    if (!childElementP)
	{
		TRACE_MID(TRACE_INFO, ("Element has no children"));
        TRACE_RETURN_TYPE(RSN_003, bool_t, FALSE);

	}

    TRACE_MID(TRACE_INFO, ("Searching child node at %.8x", childElementP));
	if (searchTreeForSQLCode(childElementP,
			                 errorMessageP,
							 sqlCodeP))
	{
		TRACE_MID(TRACE_INFO, ("DatabaseException element found"));
        TRACE_RETURN_TYPE(RSN_003, bool_t, TRUE);
	}

	/* 
	 * Now search the childs sibling elements.
	 */
    currentElementP = cniNextSibling(&rc, childElementP);
	while(currentElementP)
	{
        TRACE_MID(TRACE_INFO, ("Searching sibling node at %.8x", currentElementP));
		if (searchTreeForSQLCode(currentElementP,
			                     errorMessageP,
								 sqlCodeP))
		{
			TRACE_MID(TRACE_INFO, ("DatabaseException element found"));
            TRACE_RETURN_TYPE(RSN_003, bool_t, TRUE);
		}
	    currentElementP = cniNextSibling(&rc, currentElementP);
	}

	/* if we get to here, then we've searched the whole tree below this node, to no avail */
	TRACE_MID(TRACE_INFO, ("SQLCODE not found"));
    TRACE_RETURN_TYPE(RSN_003, bool_t, FALSE);
}


/******************************************************************************/
/* lookForDatabaseException                                                   */
/******************************************************************************/
bool_t lookForDatabaseException(IN CciElement *elementP,
						        OUT CciChar **errorMessageP,
						        OUT int *sqlCodeP)
{
	int rc;
	CciElement *insertElementP,
		       *textElementP,
		       *messageElementP = NULL;
    CciChar elementName[MAX_STRING_SIZE + 1];
    CciElementType elementType;
    CciSize nameSize;
	CciValueType valueType;
    CciChar smallBuffer[1],
		    *localBufferP = NULL;
    int numBytes,
		localBufferSize = 0;
	bool_t elementHasAName = TRUE,
		   databaseExceptionElement = FALSE,
		   sqlCodeFound;

    TRACE_ENTRY("lookForDatabaseException", ("elementP=%.8x", elementP));

	/* get the current element type */
	elementType = cniElementType(&rc, 
								 elementP);

	if (elementType != CCI_ELEMENT_TYPE_NAME &&
		elementType != CCI_ELEMENT_TYPE_NAME_VALUE)
	{
		TRACE_MID(TRACE_ERROR, ("Element has no name"));
		elementHasAName = FALSE;
	}

    if (elementHasAName)
	{
		/* get the current element name */
		nameSize = cniElementName(&rc, 
								  elementP, 
								  (CciChar *)&elementName, 
								  sizeof(elementName));

		TRACE_MID(TRACE_INFO, ("Element name is %S", elementName));

		/* Is this a DatabaseException element? */
		if (!UCcmp(elementName, C.DATABASE_EXCEPTION))
		{
			TRACE_MID(TRACE_INFO, ("DatabaseException element found"));
			databaseExceptionElement = TRUE;
		}
	}

	if (databaseExceptionElement)
	{
		/* 
		 * Search the elements of the database exception element looking for Insert 
		 * elements. 
		 */
		insertElementP = cniSearchFirstChild(&rc,
		                                     elementP,
							                 CCI_COMPARE_MODE_NAME,
							                 CCI_ELEMENT_TYPE_NAME,
								             C.INSERT);
		if (!insertElementP)
		{
			TRACE_MID(TRACE_INFO, ("No Insert found in element"));

			/* Look for any database exceptions nested in the current one */
			TRACE_MID(TRACE_INFO, ("Searching tree below database exception node"));
			sqlCodeFound = searchTreeForSQLCode(insertElementP,
									            errorMessageP,
									            sqlCodeP);
			
			TRACE_MID(TRACE_INFO, ("DatabaseException element %sfound",
				                    sqlCodeFound ? "" : "not "));
			TRACE_RETURN_TYPE(RSN_003, bool_t, sqlCodeFound);
		}

		/* check each Insert element in current Database Exception element */
		while (insertElementP)
		{
			/* look for Text values in Insert element */
			TRACE_MID(TRACE_INFO, ("Insert found in message"));

			textElementP = cniSearchFirstChild(&rc,
											 insertElementP,
											 CCI_COMPARE_MODE_NAME,
											 CCI_ELEMENT_TYPE_NAME_VALUE,
											 C.TEXT);

			while (textElementP)
			{
			    TRACE_MID(TRACE_INFO, ("Text found in message at %.8x", textElementP));

				/* extract the value of this element and check for SQLSTATE */

				/* Get the type of the element value */
	            valueType = cniElementValueType(&rc, textElementP);
				if (valueType == CCI_VALUE_TYPE_CHAR)
				{
			        TRACE_MID(TRACE_INFO, ("Element type is CCI_VALUE_TYPE_CHAR"));

					/* determine how large output buffer needs to be */
					numBytes = cniElementCharacterValue(&rc, 
														textElementP, 
														smallBuffer, 
														sizeof(smallBuffer));

					if (localBufferSize < numBytes)
					{
						if (localBufferP)
						{
							localBufferP = (CciChar *)realloc(localBufferP, 
															  numBytes * sizeof(CciChar));
					        TRACE_MID(TRACE_INFO, ("Buffer reallocated at %.8x", localBufferP));
						}
						else
						{
							localBufferP = (CciChar *)malloc(numBytes * sizeof(CciChar));
					        TRACE_MID(TRACE_INFO, ("Buffer allocated at %.8x", localBufferP));
						}

						localBufferSize = numBytes;
					}

					/* go get the data */
					numBytes = cniElementCharacterValue(&rc, 
														textElementP, 
														localBufferP, 
														localBufferSize * sizeof(CciChar));

			        TRACE_MID(TRACE_INFO, ("Extracted Text value: '%S'", localBufferP));

					/* check if text string contains SQLSTATE keyword */
					if (UCstr(localBufferP, C.SQLSTATE))
					{
						/* allocate output buffer */
						*errorMessageP = (CciChar *)malloc(localBufferSize * sizeof(CciChar));

						UCcpy(*errorMessageP, localBufferP);
						extractSQLCode(*errorMessageP, sqlCodeP);

			            TRACE_MID(TRACE_INFO, ("Found SQLSTATE"));
			            TRACE_MID(TRACE_INFO, ("ErrorMessage is: %S", *errorMessageP));
			            TRACE_MID(TRACE_INFO, ("sqlCode is: %d", *sqlCodeP));
						if (localBufferP)
						{
							TRACE_MID(TRACE_INFO, ("Free'ing buffer at %.8x", localBufferP));
							free(localBufferP);
						}
		                TRACE_RETURN_TYPE(RSN_003, bool_t, TRUE);
					}
				}
				
				textElementP = cniSearchNextSibling(&rc,
											        textElementP,
											        CCI_COMPARE_MODE_NAME,
											        CCI_ELEMENT_TYPE_NAME_VALUE,
											        C.TEXT);
	            TRACE_MID(TRACE_INFO, ("textElementP: cniSearchNextSibling returned %.8x", textElementP));

			}

			insertElementP = cniSearchNextSibling(&rc,
												 insertElementP,
												 CCI_COMPARE_MODE_NAME,
												 CCI_ELEMENT_TYPE_NAME,
												 C.INSERT);
            TRACE_MID(TRACE_INFO, ("insertElementP: cniSearchNextSibling returned %.8x", insertElementP));
		}
		/* free any locally allocated storage */
		if (localBufferP)
		{
			TRACE_MID(TRACE_INFO, ("Free'ing buffer at %.8x", localBufferP));
			free(localBufferP);
		}
	}

    TRACE_RETURN_TYPE(RSN_003, bool_t, FALSE);
}


/******************************************************************************/
/* extractSqlCode                                                             */
/******************************************************************************/
void extractSQLCode(IN CciChar *sqlErrorMessage,
					OUT int *sqlCodeP)
{
	CciChar *keywordP,
		    *sqlCodeCharP,
			valueBuffer[MAX_STRING_SIZE + 1];
	int i;

    TRACE_ENTRY("extractSQLCode", ("sqlErrorMessage=%S", sqlErrorMessage));

	/* find the start of the SQLSTATE keyword */
    keywordP = UCstr(sqlErrorMessage, C.SQLSTATE_EQUALS);
    TRACE_MID(TRACE_INFO, ("KeywordP is '%S'", keywordP));

	/* move past the SQLSTATE keyword */
	sqlCodeCharP = keywordP + UClen(C.SQLSTATE_EQUALS);
    TRACE_MID(TRACE_INFO, ("sqlCodeP is '%S'", sqlCodeCharP));

	/* extract the numeric value of SQLCODE */
	i=0;
	memset(valueBuffer, 0, sizeof(valueBuffer));
	while(UCchr(C.ALL_NUMBERS, *(sqlCodeCharP+i)))
	{
		valueBuffer[i] = *(sqlCodeCharP+i);
		i++;
	}

	*sqlCodeP = UCToInt(valueBuffer);
    TRACE_MID(TRACE_INFO, ("Extracted SQLCODE value is '%S' (%d)", 
		                    valueBuffer, *sqlCodeP));

	TRACE_RETURN_VOID(RSN_024);
}

/* local functions */
/******************************************************************************/
/* copySQLCodeToElement                                                       */
/******************************************************************************/
int copySQLCodeToElement(IN CciElement *elementP, 
					     IN int sqlCode,
					     IN bool_t newElement,
					     IN bool_t xmlMessage)
{
	int rc,
		numCciChars;
	bool_t valueAlreadyExists = FALSE;
	CciElementType elementType,
		           childElementType;
	CciElement *childElementP,
		       *dataElementP;
	CciChar sqlCodeUC[MAX_STRING_SIZE + 1];
    TRACE_ENTRY("copySQLCodeToElement", ("elementP=%.8x, sqlCode=%d, newElement=%s",
							                elementP, sqlCode,
											newElement ? "TRUE" : "FALSE",
											xmlMessage ? "TRUE" : "FALSE"));

	dataElementP = elementP;

	/* 
	 * If the element is new, then it will have been created as a NAME-VALUE element.  
	 * Simply assign it a value 
	 */
	if (newElement)
	{
		if (xmlMessage)
		{
			/* convert decimal sqlCode value to Unicode */
			intToUC(sqlCode, sqlCodeUC);

			TRACE_MID(TRACE_INFO, ("setting CHAR value for XML message"));
			numCciChars = UClen(sqlCodeUC);
			cniSetElementCharacterValue(&rc, 
										dataElementP, 
										sqlCodeUC,
										numCciChars);
		}
		else
		{
			TRACE_MID(TRACE_INFO, ("setting INTEGER value for non-XML messager"));
			cniSetElementIntegerValue(&rc, 
			                          dataElementP,
									  (CciInt)sqlCode);
		}
		if (rc)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to set SQLCODE value into specified element"));
			TRACE_RETURN(RSN_019, TT_FAILED_TO_SET_ELEMENT_VALUE);
		}
		TRACE_MID(TRACE_INFO, ("SQLCODE value successfully assigned to new element"));
		TRACE_RETURN(RSN_019, TT_SUCCESS);
	}

	/* 
	 * Check that existing element is suitable to have a value assigned to it.
	 */
	elementType = cniElementType(&rc, 
								 elementP);
	TRACE_MID(TRACE_INFO, ("elementType is %s",
		elementType == CCI_ELEMENT_TYPE_NAME ? "CCI_ELEMENT_TYPE_NAME" :
		elementType == CCI_ELEMENT_TYPE_NAME_VALUE ? "CCI_ELEMENT_TYPE_NAME_VALUE" :
		elementType == CCI_ELEMENT_TYPE_VALUE ? "CCI_ELEMENT_TYPE_VALUE" : "UNKNOWN"));

	if (elementType != CCI_ELEMENT_TYPE_NAME &&
	    elementType != CCI_ELEMENT_TYPE_NAME_VALUE &&
	    elementType != CCI_ELEMENT_TYPE_VALUE)
	{
		TRACE_MID(TRACE_ERROR, ("Element type is not suitable for storing SQLCODE value"));
		TRACE_RETURN(RSN_019, TT_INVALID_ELEMENT_TYPE_SPECIFIED);
	}

	if (elementType == CCI_ELEMENT_TYPE_NAME)
	{
		/* address the value element if it exists, otherwise create one */
		childElementP = cniFirstChild(&rc,
                                      elementP);
		if (childElementP)
		{
			/* is it a NameValue type? */
			childElementType = cniElementType(&rc, 
											  childElementP);
			if (childElementType != CCI_ELEMENT_TYPE_VALUE)
			{
				TRACE_MID(TRACE_ERROR, ("Element type is not suitable for storing SQLCODE value"));
				TRACE_RETURN(RSN_019, TT_INVALID_ELEMENT_TYPE_SPECIFIED);
			}
			valueAlreadyExists = TRUE;
		}
		else
		{
			// create a child VALUE element
			childElementP = createChildElement(elementP,
				                               NULL,
											   CCI_ELEMENT_TYPE_VALUE);
		}
		dataElementP = childElementP;
	}
	else
	{
		valueAlreadyExists = TRUE;
		dataElementP = elementP;
	}

    /* 
	 * If new element created, then just copy in the value.
	 */
	if (!valueAlreadyExists || !xmlMessage)
	{
		TRACE_MID(TRACE_INFO, ("setting INTEGER value for non-XML messager"));
		cniSetElementIntegerValue(&rc, 
			                      dataElementP,
								  (CciInt)sqlCode);
	}
	else
	{
		/* convert decimal sqlCode value to Unicode */
		intToUC(sqlCode, sqlCodeUC);

		TRACE_MID(TRACE_INFO, ("setting CHAR value for XML message"));
		numCciChars = UClen(sqlCodeUC);
		cniSetElementCharacterValue(&rc, 
									dataElementP, 
									sqlCodeUC,
									numCciChars);
	}

	if (rc)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set SQLCODE value into specified element"));
		TRACE_RETURN(RSN_019, TT_FAILED_TO_SET_ELEMENT_VALUE);
	}

	TRACE_MID(TRACE_INFO, ("SQLCODE value successfully assigned to new element"));
	TRACE_RETURN(RSN_019, TT_SUCCESS);
}
