#if !defined(TRACE_DEFS_H)
#define TRACE_DEFS_H

/***************************************************************************
 * Trace macro's
 *
 * The set of macros below provide the backbone of the tracing function.
 *
 * When tracing is enabled each macro will output the requested trace to 
 * the selected destination. 
 *
 *
 * Function entry macros
 *
 * TRACE_ENTRY - Is called at the start of each function for which trace will
 * be required. It accepts as parameters a function name and a set of
 * parameters enclosed within brackets '(', ')' which are suitable for
 * output using printf.
 *
 *
 * Function Return macros
 *
 * There are 3 macros which replace the return statement for functions.
 * The 3 macros handle void functions, functions which return integers
 * and functions which return other types.
 * They all share a common 1st paramter which is a return sequence number (RSN).
 * The purpose of the RSN, is to allow the point at which the function 
 * returned/exited, to be uniquely identified during debugging.
 *
 * The RSN's are defined within the trace header file and must be unique 
 * within that source file, to allow the user debugging to quickly locate
 * the exit point from he function.
 *    eg.
 *    #define RSN_001 1
 *    #define RSN_002 2
 *    #define RSN_003 3
 *    #define RSN_004 4
 *    #define RSN_005 5
 *
 * 
 * TRACE_RETURN_VOID - replaces the 'return;' void return call normally used 
 * to return from a void function. It single parameter is an integer return
 * sequence number which will be written to the trace and is used to 
 * identify the return point used by a function from the trace.
 *     eg.
 *     TRACE_RETURN_VOID(RSN_001);
 *
 * TRACE_RETURN - replaces the standard return call from a function, where 
 * the function is returning an 'int'. Its first parameter is an integer
 * return sequence number which will be written to the trace and is used to
 * identify the return point used by a function from the trace. The second
 * parameter is the value to be returned from the function.
 *     eg.
 *     TRACE_RETURN(RSN_002, retCode);
 *
 * TRACE_RETURN_TYPE - replaces the return statement from a function where the 
 * function returns a value which is not equivalent to an int. Like the 
 * previous 2 return macros, it takes a return sequence number as its first 
 * parameter, followed by a typecast variable used to identify the type of the 
 * return and the value to be returned.
 *     eg.
 *     TRACE_RETURN_TYPE(RSN_003, struct MyStruct *, myPointer);
 *     TRACE_RETURN_TYPE(RSN_004, char, myChar);
 *
 * In addition to the 3 return macros, there are 3 further macros which can be
 * used to generate additional trace.
 *
 * TRACE_EXIT - replaces the exit statement.  It takes a return sequence 
 * number as its first parameter followed by the exit value which should 
 * resolve to an integer type.
 *     eg.
 *     TRACE_EXIT(RSN_009, 5);
 *
 * TRACE_MID - is used to generate intermediate trace entries for significant
 * events during the execution of a program.  It takes 2 parameters. The first
 * parameter is the entrys trace level.  This determines the trace level(s) for
 * which the entry will be output.  The second parameter is a set of
 * parameters enclosed within brackets '(', ')' which are suitable for output
 * using printf.
 *     eg.
 *     TRACE_MID(TRACE_INFO, ("Interim value of counter is %d", mycount));
 *
 * TRACE_MIDHEX - is used to generate intermediate trace entries containing
 * hexadecimal representation of data items.  It takes 3 parameters.  The first
 * parameter is the entry's trace level.  This determines the trace level(s) for
 * which the entry will be output.  The subsequent parameters are the address
 * and length of the data area to be displayed in the trace.
 *     eg.
 *     TRACE_MIDHEX(TRACE_INFO, myDataArea, sizeof(myDataArea));
 *
 ***************************************************************************/

#define MAX_FUNCTION_NAME          100
#define MAX_ERROR_MESSAGE_LENGTH 4096

#if !defined(bool_t)
    #define bool_t int
#endif
 
#define TRACE_VARS \
    char __trace_functionName[MAX_FUNCTION_NAME + 1];

#if defined(BUILTIN_TRACE)
#define TRACE_ENTRY_WITHOUT_VARS(__inputFunctionName, __inputParameters) \
    strncpy(__trace_functionName, __inputFunctionName, MAX_FUNCTION_NAME); \
    __trace_functionName[MAX_FUNCTION_NAME] = '\0'; \
    if (!traceInitialised) \
    {\
	    TraceInit(); \
    }\
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceEntry(__trace_functionName); \
	    TraceEntryArgs __inputParameters;\
    }
#else
#define TRACE_ENTRY_WITHOUT_VARS(__inputFunctionName, __inputParameters) \
    strncpy(__trace_functionName, __inputFunctionName, MAX_FUNCTION_NAME); \
    __trace_functionName[MAX_FUNCTION_NAME] = '\0'; \
    if (firstTraceCall) \
    {\
	    TraceInit(traceMaxRetcodeValue, traceRetcodeString, \
		          traceEnvironmentVariable, &traceCallerIndex); \
		firstTraceCall = FALSE; \
    }\
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceEntry(__trace_functionName); \
	    TraceEntryArgs __inputParameters;\
    } 
#endif

#if defined(BUILTIN_TRACE)
#define TRACE_ENTRY(__inputFunctionName, __inputParameters) \
    char __trace_functionName[MAX_FUNCTION_NAME + 1]; \
    strncpy(__trace_functionName, __inputFunctionName, MAX_FUNCTION_NAME); \
    __trace_functionName[MAX_FUNCTION_NAME] = '\0'; \
    if (!traceInitialised) \
    { \
	    TraceInit(); \
    } \
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceEntry(__trace_functionName); \
	    TraceEntryArgs __inputParameters;\
    }
#else
#define TRACE_ENTRY(__inputFunctionName, __inputParameters) \
    char __trace_functionName[MAX_FUNCTION_NAME + 1]; \
    strncpy(__trace_functionName, __inputFunctionName, MAX_FUNCTION_NAME); \
    __trace_functionName[MAX_FUNCTION_NAME] = '\0'; \
    if (firstTraceCall) \
    { \
	    TraceInit(traceMaxRetcodeValue, traceRetcodeString, \
		          traceEnvironmentVariable, &traceCallerIndex); \
	    firstTraceCall = FALSE; \
    } \
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceEntry(__trace_functionName); \
	    TraceEntryArgs __inputParameters;\
    }

#endif

#define TRACE_RETURN_VOID(_RSN) \
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceReturnVoid(__trace_functionName, _RSN); \
    } \
    return;

#if defined(BUILTIN_TRACE)
#define TRACE_RETURN(_RSN, _retcode) \
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceReturnInteger(__trace_functionName, _RSN, _retcode); \
    } \
    return(_retcode);
#else
#define TRACE_RETURN(_RSN, _retcode) \
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceReturnInteger(__trace_functionName, _RSN, _retcode, traceCallerIndex); \
    } \
    return(_retcode);
#endif

#define TRACE_RETURN_TYPE(_RSN, _type, _returnValue) \
{ \
	_type _TRACE_RETURNVALUE=(_returnValue); \
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceReturnType(__trace_functionName, _RSN, \
                        #_type, (void *)&_TRACE_RETURNVALUE); \
    } \
	return(_returnValue); \
}

#define TRACE_EXIT(_RSN, _exitValue) \
    if (traceEnabled && (traceFlags & TRACE_ENTRY_EXIT)) \
    { \
	    TraceExit(__trace_functionName, _RSN, _exitValue); \
    } \
    exit(_exitValue);

#if defined(BUILTIN_TRACE)
#define TRACE_MID(_traceLevel, _traceString) \
    if (!traceInitialised) \
    {\
	    TraceInit(); \
    }\
    if (traceEnabled && (traceFlags & _traceLevel)) \
    { \
	    TraceMidArgs _traceString; \
    }
#else
#define TRACE_MID(_traceLevel, _traceString) \
    if (!traceInitialised) \
    {\
	    TraceInit(traceMaxRetcodeValue, traceRetcodeString, traceEnvironmentVariable, \
		          &traceCallerIndex); \
    }\
    if (traceEnabled && (traceFlags & _traceLevel)) \
    { \
	    TraceMidArgs _traceString; \
    }
#endif

#if defined(BUILTIN_TRACE)
#define TRACE_MIDHEX(_traceLevel, _traceBuffer, _traceLength) \
    if (!traceInitialised) \
    {\
	    TraceInit(); \
    }\
    if (traceEnabled && (traceFlags & _traceLevel)) \
    { \
	    TraceHexDump (_traceBuffer, _traceLength); \
    }
#else
#define TRACE_MIDHEX(_traceLevel, _traceBuffer, _traceLength) \
    if (!traceInitialised) \
    {\
	    TraceInit(traceMaxRetcodeValue, traceRetcodeString, \
		          traceEnvironmentVariable, &traceCallerIndex); \
    }\
    if (traceEnabled && (traceFlags & _traceLevel)) \
    { \
	    TraceHexDump (_traceBuffer, _traceLength); \
    }
#endif

#define GEN_ERROR_MESSAGE(__errMsg) \
    TraceGenErrorMessage __errMsg;  \
    TRACE_MID(TRACE_ERROR, __errMsg );


/* 
 * defines controlling generation of function prototypes and definitions
 */
#if defined(WIN32)
	#if !defined(BUILTIN_TRACE)
	    #if defined(DLLEXPORT_TRACE)
		#if defined(TR_LIBSPEC)
		    #undef TR_LIBSPEC
		#endif
		#define TR_LIBSPEC	__declspec( dllexport )
	    #else
		#if defined(TR_LIBSPEC)
		    #undef TR_LIBSPEC
		#endif
		#define TR_LIBSPEC __declspec( dllimport )
	    #endif
	    
	    #if !defined(TR_CALLCONV)
		#define TR_CALLCONV __stdcall
	    #endif
	#else
	    #if defined(TR_LIBSPEC)
		#undef TR_LIBSPEC
	    #endif
	    #define TR_LIBSPEC	
	    
	    #if defined(TR_CALLCONV)
		#undef TR_CALLCONV
	    #endif
	    #define TR_CALLCONV
	#endif
#else
    #define TR_LIBSPEC	
    #define TR_CALLCONV
#endif

#if !defined(IN)
    #define IN
#endif
#if !defined(OUT)
    #define OUT
#endif
#if !defined(INOUT)
    #define INOUT
#endif

/* supported trace settings */
#define TRACE_NONE                0
#define TRACE_ENTRY_EXIT       1<<1
#define TRACE_ERROR            1<<2
#define TRACE_INFO             1<<3
#define TRACE_WARNING          1<<4
#define TRACE_USER_1           1<<5
#define TRACE_USER_2           1<<6
#define TRACE_USER_3           1<<7
#define TRACE_USER_4           1<<8
#define TRACE_USER_5           1<<9
#define TRACE_USER_6           1<<10
#define TRACE_USER_7           1<<11
#define TRACE_USER_8           1<<12
#define TRACE_USER_9           1<<13
#define TRACE_USER_10          1<<14
#define TRACE_USER_1_TO_4 (TRACE_USER_1|TRACE_USER_2|TRACE_USER_3|TRACE_USER_4)
#define TRACE_USER_5_TO_8 (TRACE_USER_5|TRACE_USER_6|TRACE_USER_7|TRACE_USER_8)
#define TRACE_USER_9_TO_10 (TRACE_USER_9|TRACE_USER_10)
#define TRACE_USER_ALL (TRACE_USER_1_TO_4|TRACE_USER_5_TO_8|TRACE_USER_9_TO_10)
#define TRACE_SYSTEM_ALL (TRACE_ENTRY_EXIT|TRACE_ERROR|TRACE_INFO|TRACE_WARNING)
#define TRACE_ALL (TRACE_SYSTEM_ALL | TRACE_USER_ALL)

#endif
