#define DECLARE_UTILITIES_DATA 1
/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : utilities.c                                              */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : A set of helper functions designed for use within MQSI   */
/*                   plug-in nodes.                                           */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Feb 2000                       */
/*                               Revised Date: Mar 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "utilities.h"      /* Specific header for this dll                       */
#include "unicode.h"
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */

typedef struct tagEntry {
	char *tagName;
	bool_t containsNameValueChild,
	       containsValueChild;
	struct tagEntry *nextP,
		            *prevP;
} tagEntry_t;
#define TAG_ENTRY_DEFAULT \
	NULL,    /* tagName */ \
    FALSE,   /* containsNameValueChild */ \
    FALSE,   /* containsValueChild */ \
    NULL,    /* nextP */   \
	NULL     /* prevP */

/* local function prototypes */
int processXMLTagEntry(IN char **XMLP,
				       IN int *XMLLengthP,
				       IN char **currentXMLP,
				       tagEntry_t *tagEntryP);
int processXMLTagList(IN char **XMLP,
				 	  IN int *XMLLengthP,
					  IN char **currentXMLP,
					  tagEntry_t *tagEntryAnchorP);


/******************************************************************************/
/* updateMsgIdIndex                                                           */
/******************************************************************************/
msgIdIndexElement_t *updateMsgIdIndex(IN char *messageId,
									  IN msgIdIndexElement_t *indexAnchorP)
{
	msgIdIndexElement_t *newIndexP,
		                *currentIndexP;

    TRACE_ENTRY("updateMsgIdIndex", ("messageId=%.8x", messageId));

	TRACE_MID(TRACE_INFO, ("New message index value is..."));
	TRACE_MIDHEX(TRACE_INFO, messageId, MQ_MSG_ID_LENGTH);

	/* add entry to message id index */
	TRACE_MID(TRACE_INFO, ("About to allocate storage for new message index element"));
	newIndexP = (msgIdIndexElement_t *)malloc(sizeof(msgIdIndexElement_t));
	TRACE_MID(TRACE_INFO, ("Allocated new index element at %.8x", newIndexP));
	TRACE_ALLOC("utilities.c-40", newIndexP);
	memset(newIndexP, 0, sizeof(msgIdIndexElement_t));

    memcpy(newIndexP->msgId, messageId, MQ_MSG_ID_LENGTH);

	currentIndexP = indexAnchorP;

	/* insert index element into chain in message id order */
	while(currentIndexP)
	{
		TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentIndexP));
		if (!currentIndexP->nextP ||
			memcmp(messageId, currentIndexP->nextP->msgId, MQ_MSG_ID_LENGTH) < 0)
		{
		    TRACE_MID(TRACE_INFO, ("Inserting new element"));
			break;
		}
		currentIndexP = currentIndexP->nextP;
	}

    INSERT_AFTER_IN_LINKED_LIST(currentIndexP, newIndexP);

	TRACE_RETURN_TYPE(RSN_001, msgIdIndexElement_t *, newIndexP);
}

/******************************************************************************/
/* findMsgIndexElement                                                           */
/******************************************************************************/
msgIdIndexElement_t *findMsgIndexElement(IN char *id,
									     IN msgIdIndexElement_t *indexAnchorP)
{
	msgIdIndexElement_t *currentIndexP;


    TRACE_ENTRY("findMsgIndexElement", ("id=%.8x, indexAnchorP=%.8x", 
		                                 id, indexAnchorP));

	currentIndexP = indexAnchorP->nextP;

	/* find matching index value in chain */
	while(currentIndexP)
	{
		TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentIndexP));
		if (!memcmp(id, currentIndexP->msgId, MQ_MSG_ID_LENGTH))
		{
		    TRACE_MID(TRACE_INFO, ("Found a match"));
			TRACE_RETURN_TYPE(RSN_002, msgIdIndexElement_t *, currentIndexP);
		}

		/* check if we've gone past the value we're looking for */
		if (memcmp(id, currentIndexP->msgId, MQ_MSG_ID_LENGTH) < 0)
		{
		    TRACE_MID(TRACE_INFO, ("Gone past value being searched for"));
			TRACE_RETURN_TYPE(RSN_003, msgIdIndexElement_t *, NULL);
		}

		currentIndexP = currentIndexP->nextP;
	}

	TRACE_RETURN_TYPE(RSN_004, msgIdIndexElement_t *, NULL);
}

/******************************************************************************/
/* findMessageFlow                                                            */
/******************************************************************************/
messageFlow_t *findMessageFlow(IN CciChar *name,
							   IN messageFlow_t *msgFlowAnchorP)
{
	messageFlow_t *mfP;

    TRACE_ENTRY("findMessageFlow", ("name=%S", name));

	mfP = msgFlowAnchorP->nextP;

	/* find matching flow name in chain */
	while(mfP)
	{
		TRACE_MID(TRACE_INFO, ("Processing message flow entry at %.8x", mfP));
		if (!UCcmp(name, mfP->name))
		{
		    TRACE_MID(TRACE_INFO, ("Found a match"));
			TRACE_RETURN_TYPE(RSN_002, messageFlow_t *, mfP);
		}

		mfP = mfP->nextP;
	}

	TRACE_RETURN_TYPE(RSN_004, messageFlow_t *, NULL);
}

/******************************************************************************/
/* getMessageCharArrayValue                                                   */
/******************************************************************************/
int getMessageCharArrayValue(IN void *message,
							 IN matchType_t matchType,
							 OUT CciChar *buffer,
							 INOUT int *bufferLengthP,
							 OUT void **elementPP)
{
	void *messageElementP;
    struct CciByteArray byteArrayDetails;
    CciSize numBytes;
	CciChar unicodeName[MQ_Q_NAME_LENGTH + 1];
	int rc;

    TRACE_ENTRY("getMessageCharArrayValue", ("message=%.8x, matchType=%s, "
		                                     "buffer=%.8x, *bufferLengthP=%d",
		                                     message, 
											 matchType==MESSAGE_ID ? "MESSAGE_ID" :
											 matchType==CORREL_ID ? "CORREL_ID" :
											 matchType==REPLY_TO_QUEUE ? "REPLY_TO_QUEUE" :
											 matchType==REPLY_TO_QUEUE_MANAGER ? 
                                              "REPLY_TO_QUEUE_MANAGER" : "UNKNOWN",
											  buffer, bufferLengthP ? *bufferLengthP :
											  0));
	/* access the required element */
	messageElementP = getMessageElement(message, matchType);

	if (!messageElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Message element not found in message"));
		TRACE_RETURN(RSN_005, TT_ELEMENT_NOT_FOUND);
	}

	if (matchType == MESSAGE_ID ||
	    matchType == CORREL_ID)
	{
		/* extract the element value */
		TRACE_MID(TRACE_INFO, ("Calling cniElementByteArrayValue"));
		byteArrayDetails.pointer = (void *)buffer;
		byteArrayDetails.size = *bufferLengthP;
		numBytes = cniElementByteArrayValue(&rc,
											messageElementP,
											&byteArrayDetails);
		if (rc == CCI_SUCCESS)
		{
			*bufferLengthP = numBytes;
		}
	}
	else
	if (matchType == REPLY_TO_QUEUE ||
	    matchType == REPLY_TO_QUEUE_MANAGER)
	{
		/* character values in MQMD are stored in unicode format */
		TRACE_MID(TRACE_INFO, ("Calling cniElementCharacterValue"));
		memset(unicodeName, 0, sizeof(unicodeName));
		numBytes = cniElementCharacterValue(&rc,
			                                messageElementP,
								            buffer,
											*bufferLengthP);

	}
	else
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Unknown data type specified, matchType = %d",
			                    matchType));
		TRACE_RETURN(RSN_007, TT_INVALID_ELEMENT_TYPE_SPECIFIED);
	}

	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract element value from message (rc=%d)",
								rc));
		TRACE_RETURN(RSN_008, TT_ELEMENT_VALUE_NOT_RETRIEVED);
	}

	/* return pointer to retrieved element if user requested it */
	if (elementPP)
	{
		TRACE_MID(TRACE_INFO, ("Returning element at %.8x", messageElementP));
		*elementPP = messageElementP;
	}
	TRACE_MID(TRACE_INFO, ("Element value retrieved, length is %d, value is...", 
		                   *bufferLengthP));
	TRACE_MIDHEX(TRACE_INFO, buffer, *bufferLengthP);

	TRACE_RETURN(RSN_009, TT_SUCCESS);
}

/******************************************************************************/
/* createNewMessage                                                           */
/******************************************************************************/
CciMessage *createNewMessage(IN CciMessage *message)
{
	CciMessage *newMsg;
	CciElement *newMsgRootP,
		       *oldMsgRootP;
	int rc;
	CciMessageContext *msgContextP;

    TRACE_ENTRY("createNewMessage", ("message=%.8x", message));

	/* create a new message, using current message as a template */
	msgContextP = cniGetMessageContext(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get message context "
			                    "(cniGetMessageContext rc=%d)", rc));
		TRACE_RETURN_TYPE(RSN_010, void *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

	/* Create a new message */
    newMsg = cniCreateMessage(&rc, msgContextP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new message (cniCreateMessage FAILED)"));
		TRACE_RETURN_TYPE(RSN_011, void *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

	/* Get the root element of new message */
    newMsgRootP = cniRootElement(&rc, newMsg);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("cniRootElement for new message FAILED"));
		TRACE_RETURN_TYPE(RSN_012, void *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("Addressed new message root element at %.8x", 
		                    newMsgRootP));

	/* Get the root element of original message */
    oldMsgRootP = cniRootElement(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("cniRootElement for original message FAILED"));
		TRACE_RETURN_TYPE(RSN_013, void *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("Addressed original message root element at %.8x", 
		                    oldMsgRootP));

	/* copy original message to new message tree */
	cniCopyElementTree(&rc,
                       oldMsgRootP,
			           newMsgRootP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy original message to new message"));
		TRACE_RETURN_TYPE(RSN_014, void *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("New message successfully created from original"));


	/* return the new message to caller */
	TRACE_RETURN_TYPE(RSN_015, void *, newMsg);
}

/******************************************************************************/
/* findChild                                                                  */
/******************************************************************************/
void *findChild(void *currentElementP,
				   CciChar *elementName)
{
	int rc;
	void *localElementP = currentElementP;
	CciSize elementNameSize;
	CciChar localElementName[256];

    TRACE_ENTRY("findChild", ("currentElementP=%.8x, elementName=%S",
							 currentElementP, elementName));

	while (localElementP)
	{
        elementNameSize = cniElementName(&rc, 
		                                 localElementP, 
									     (CciChar*)&localElementName, 
									     sizeof(localElementName));
        TRACE_MID(TRACE_INFO, ("Checking element name '%S'", localElementName));

		if (!UCcmp(elementName, localElementName))
		{
			/* found the MQMD element */
	        TRACE_RETURN_TYPE(RSN_016, void *, localElementP);
		}
        currentElementP = cniNextSibling(&rc, 
			                             currentElementP);
	}

	/* return NULL */
	TRACE_RETURN_TYPE(RSN_017, void *, NULL);
}

/******************************************************************************/
/* findElementInTree                                                          */
/******************************************************************************/
void *findElementInTree(void *currentElementP,
				           CciChar *elementName)
{
	int rc;
	void *localElementP = currentElementP,
		 *currentSiblingP = currentElementP;
	CciSize elementNameSize;
	CciChar localElementName[256];

    TRACE_ENTRY("findElementInTree", ("currentElementP=%.8x, elementName=%S",
							           currentElementP, elementName));

	while (localElementP)
	{
		/* does current node match? */
        elementNameSize = cniElementName(&rc, 
		                                 localElementP, 
									     (CciChar*)&localElementName, 
									     sizeof(localElementName));
        TRACE_MID(TRACE_INFO, ("Checking element name '%S'", localElementName));

		if (!UCcmp(elementName, localElementName))
		{
			/* found the MQMD element */
	        TRACE_RETURN_TYPE(RSN_018, void *, localElementP);
		}

		/* check for child under current node */
        localElementP = findChild(localElementP,
			                      elementName);
		if (localElementP)
		{
			break;
		}

		/* get next sibling */
        currentSiblingP = cniNextSibling(&rc, 
			                             currentSiblingP);
		localElementP = currentSiblingP;
        TRACE_MID(TRACE_INFO, ("Next sibling at %.8x", localElementP));
	}

	TRACE_RETURN_TYPE(RSN_019, CciChar *, localElementP);
}

/******************************************************************************/
/* getMessageElement                                                          */
/******************************************************************************/
void *getMessageElement(IN void *message,
					    IN matchType_t matchType)
{
	int rc;
	void *rootElementP,
		 *MQMDP,
		 *destinationP,
		 *MQDestinationListP,
		 *DestinationDataP,
		 *messageElementP = NULL;
    CciChar elementName[256];
    CciSize elementNameSize;

    TRACE_ENTRY("getMessageElement", ("message=%.8x, matchType=%s", 
		                               message,
									   matchType==BODY ? "BODY" :
									   matchType==MQMD_ELEMENT ? "MQMD" :
									   matchType==MESSAGE_ID ? "MESSAGE_ID" :
	                                   matchType==CORREL_ID ? "CORREL_ID" :
	                                   matchType==REPLY_TO_QUEUE ? "REPLY_TO_QUEUE" :
                           matchType==REPLY_TO_QUEUE_MANAGER ? "REPLY_TO_QUEUE_MANAGER" :
                           matchType==DESTINATION ? "DESTINATION" :
                           matchType==MQ_DESTINATION_LIST ? "MQ_DESTINATION_LIST" :
                           matchType==DESTINATION_DATA ? "DESTINATION_DATA" :
                           matchType==PUT_APPL_NAME ? "PUT_APPL_NAME" :
	                                   "unknown"));

	/* Get the root element */
    rootElementP = cniRootElement(&rc, message);

	if (matchType == DESTINATION || 
		matchType == MQ_DESTINATION_LIST ||
		matchType == DESTINATION_DATA)
	{

		/* assume we're processing a destination list message structure */
		destinationP = cniSearchFirstChild(&rc,
		                                   rootElementP,
							               CCI_COMPARE_MODE_NAME,
							               CCI_ELEMENT_TYPE_NAME,
								           U.DESTINATION);
		if (!destinationP)
		{
			/* Destination not found in message - generate error */
			TRACE_MID(TRACE_ERROR, ("Destination not found in message"));
			TRACE_RETURN_TYPE(RSN_020, void *, NULL);
		}

	    TRACE_MID(TRACE_INFO, ("Destination element located at %.8x", destinationP));

		if (matchType == DESTINATION)
		{
			TRACE_RETURN_TYPE(RSN_021, void *, destinationP);
		}

		/* address the MQDestinationList node */
		MQDestinationListP = cniSearchFirstChild(&rc,
		                                         destinationP,
											     CCI_COMPARE_MODE_NAME,
											     CCI_ELEMENT_TYPE_NAME,
								                 U.MQ_DESTINATION_LIST);
		if (!MQDestinationListP)
		{
			/* MQDestinationList not found in message - generate error */
			TRACE_MID(TRACE_ERROR, ("MQDestinationList not found in message"));
			TRACE_RETURN_TYPE(RSN_022, void *, NULL);
		}

	    TRACE_MID(TRACE_INFO, ("MQDestinationList element located at %.8x", 
			                    MQDestinationListP));

		if (matchType == MQ_DESTINATION_LIST)
		{
			TRACE_RETURN_TYPE(RSN_023, void *, MQDestinationListP);
		}

		/* address the DestinationData node(s) */
		DestinationDataP = cniSearchFirstChild(&rc,
		                                         MQDestinationListP,
											     CCI_COMPARE_MODE_NAME,
											     CCI_ELEMENT_TYPE_NAME,
								                 U.DESTINATION_DATA);
		if (!DestinationDataP)
		{
			/* DestinationData not found in message - generate error */
			TRACE_MID(TRACE_ERROR, ("DestinationData not found in message"));
			TRACE_RETURN_TYPE(RSN_024, void *, NULL);
		}

	    TRACE_MID(TRACE_INFO, ("First DestinationData element located at %.8x", 
			                    DestinationDataP));
		{
			void *nextddP,
				 *currentddP;
			/* iterate through each DestinationData instance */
			currentddP = DestinationDataP;
			nextddP = cniSearchNextSibling(&rc,
		                                   currentddP,
										   CCI_COMPARE_MODE_NAME,
										   CCI_ELEMENT_TYPE_NAME,
								           U.DESTINATION_DATA);
			while (nextddP)
			{
				TRACE_MID(TRACE_INFO, ("Next DestinationData element located at %.8x", 
										nextddP));
				currentddP = nextddP;
				nextddP = cniSearchNextSibling(&rc,
											   currentddP,
											   CCI_COMPARE_MODE_NAME,
											   CCI_ELEMENT_TYPE_NAME,
											   U.DESTINATION_DATA);
			}
		}

		if (matchType == DESTINATION_DATA)
		{
			TRACE_RETURN_TYPE(RSN_025, void *, DestinationDataP);
		}

		TRACE_RETURN_TYPE(RSN_026, void *, NULL);
		
	}

	/* 
	 * This code needs modifying so that it can correctly detect the start of the
	 * body.  This may not be the last child of root, e.g. message with 0 length data
	 */
	if (matchType == BODY)
	{
	    /* Return the last child of a target element */
		messageElementP = cniLastChild(&rc,
                                       rootElementP);
		if (!messageElementP)
		{
			/* body not found in message - generate error */
			TRACE_MID(TRACE_ERROR, ("Failed to address last child "
									"(cniLastChild rc=%d)", rc));
		    TRACE_RETURN_TYPE(RSN_027, void *, NULL);
		}

		/*
		 * check that element name is BODY - this is required so that we can 
		 * detect messages which have no body, i.e. message with length = 0 
		 */

		/* Get the element's name */
		TRACE_MID(TRACE_INFO, ("calling cniElementName for element %.8x", 
			                   messageElementP));
		elementNameSize = cniElementName(&rc, 
										 messageElementP, 
										 (CciChar *)&elementName, 
										 sizeof(elementName));

		TRACE_MID(TRACE_INFO, ("elementNameSize is %d", elementNameSize));

		/* empty wc string seems to have a length of 1 */
		if (elementNameSize <= 1)
		{
			TRACE_MID(TRACE_ERROR, ("Element has no name"));
		    TRACE_RETURN_TYPE(RSN_028, void *, NULL);
		}

		TRACE_MID(TRACE_INFO, ("element name is %S", elementName));

		TRACE_RETURN_TYPE(RSN_029, void *, messageElementP);
	}

	MQMDP = cniSearchFirstChild(&rc,
		                        rootElementP,
							    CCI_COMPARE_MODE_NAME,
							    CCI_ELEMENT_TYPE_NAME,
								U.MQMD);
    if (!MQMDP)
	{
		/* MQMD not found in message - generate error */
        TRACE_MID(TRACE_ERROR, ("MQMD not found in message"));
	    TRACE_RETURN_TYPE(RSN_030, void *, messageElementP);
	}
    TRACE_MID(TRACE_INFO, ("MQMD element located at %.8x", MQMDP));

	if (matchType == MQMD_ELEMENT)
	{
		TRACE_RETURN_TYPE(RSN_031, void *, MQMDP);
	}

	/* find required element in MQMD */
	messageElementP = cniSearchFirstChild(&rc,
		                             MQMDP,
							         CCI_COMPARE_MODE_NAME,
							         CCI_ELEMENT_TYPE_NAME_VALUE,
									 matchType==MESSAGE_ID ? U.MSG_ID :
									 matchType==CORREL_ID ? U.CORREL_ID :
									 matchType==PUT_APPL_NAME ? U.PUT_APPL_NAME :
									 matchType==REPLY_TO_QUEUE ? U.REPLY_TO_Q :
									 matchType==REPLY_TO_QUEUE_MANAGER ? U.REPLY_TO_Q_MGR : U.NULL_STRING);

    if (!messageElementP)
	{
		/* Required element not found in MQMD - generate error */
        TRACE_MID(TRACE_ERROR, ("%s not found in MQMD",
							   matchType==MESSAGE_ID ? "Message Id" :
							   matchType==CORREL_ID ? "Correl Id" :
							   matchType==PUT_APPL_NAME ? "PutApplName" :
							   matchType==REPLY_TO_QUEUE ? "ReplyToQ" :
							   matchType==REPLY_TO_QUEUE_MANAGER ? "ReplyToQMgr" : "?"));
	    TRACE_RETURN_TYPE(RSN_032, void *, messageElementP);
	}
    TRACE_MID(TRACE_INFO, ("%s element located at %.8x", 
						   matchType==MESSAGE_ID ? "MsgId" :
						   matchType==CORREL_ID ? "CorrelId" :
						   matchType==PUT_APPL_NAME ? "PutApplName" :
						   matchType==REPLY_TO_QUEUE ? "ReplyToQ" :
						   matchType==REPLY_TO_QUEUE_MANAGER ? "ReplyToQMgr" : "?",
						   messageElementP));

	/* return to caller */
	TRACE_RETURN_TYPE(RSN_033, void *, messageElementP);
}

/******************************************************************************/
/* propagateMessage                                                           */
/******************************************************************************/
int propagateMessage(IN context_t *contextP,
					 IN CciChar *terminalName,
					 IN void *message,
					 IN void *destinationList,
					 IN void *exceptionList,
					 IN bool_t newMessage)	
{
	int retValue,
		  rc;
	void *outputTerminalObject;
    CCI_EXCEPTION_ST exceptionStructure;
	char functionName[] = "propagateMessage()",
		 msgBuffer[1000],
		 sbTerminalName[MAX_TERMINAL_NAME_LEN],
		 rcString[20];

    TRACE_ENTRY("propagateMessage", ("contextP=%.8x, terminalName=%S, "
		                     "message=%.8x, newMessage=%s",
							 contextP, terminalName, message,
							 newMessage ? "TRUE" : "FALSE"));

    /* get the handle to the output terminal */
    outputTerminalObject = getOutputTerminalHandle(contextP, terminalName);

    /* check to see if terminal exists */
    if (!outputTerminalObject)
	{
	    UCsprintf(msgBuffer, "Requested terminal %S does not exist in message flow",
			               terminalName);
		UCsprintf(sbTerminalName, "%S", terminalName);

        /* Terminal does not exist */
        cciLog(&rc, 
		       CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_TERMINAL_DOES_NOT_EXIST, 
			   msgBuffer,
#if !defined(WIN32)
		       brokerName,
#endif
			   sbTerminalName,
			   (char *)0);

	   TRACE_RETURN(RSN_034, TT_TERMINAL_DOES_NOT_EXIST);
	}

    /* check to see if terminal is attached */
    if (!cniIsTerminalAttached(&rc, outputTerminalObject))
	{
	   /* Terminal not attached */
	    UCsprintf(msgBuffer, "Requested terminal %S is not attached in message flow",
			               terminalName);
		UCsprintf(sbTerminalName, "%S", terminalName);

        /* Terminal does not exist */
        cciLog(&rc, 
		       CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_TERMINAL_NOT_ATTACHED, 
			   msgBuffer,
#if !defined(WIN32)
		       brokerName,
#endif
			   sbTerminalName,
			   (char *)0);

		TRACE_RETURN(RSN_035, TT_TERMINAL_NOT_ATTACHED);
	}

    if (rc == CCI_SUCCESS)
	{
	    TRACE_MID(TRACE_INFO, ("About to call cniPropagate"));
        retValue = cniPropagate(&rc,
					            outputTerminalObject,
					            destinationList,
					            exceptionList,
					            message);
	    TRACE_MID(TRACE_INFO, ("Returned from call to cniPropagate"));
		if (rc != CCI_SUCCESS)
		{
		    if (rc == CCI_EXCEPTION)
			{
	            TRACE_MID(TRACE_ERROR, ("Caught an exception"));

				/* initialise the exception structure...requires more than memset!! */
				memset(&exceptionStructure, 0, sizeof(exceptionStructure));

				cciGetLastExceptionData(&rc, &exceptionStructure);
	            TRACE_MID(TRACE_ERROR, ("Exception text is '%s'",
					                     exceptionStructure.traceText));
				TRACE_RETURN(RSN_035, TT_PROPAGATION_EXCEPTION);
			}

			sprintf(msgBuffer, "Failed to propagate message (rc=%d)", rc);
			sprintf(rcString, "%d", rc);
			/* log error message */
			cciLog(&rc, 
				   CCI_LOG_ERROR, 
				   __FILE__, 
				   __LINE__, 
				   functionName, 
				   C.PLUGIN_MESSAGE_CATALOGUE, 
				   MJB_MESSAGE_NOT_PROPAGATED, 
				   msgBuffer,
#if !defined(WIN32)
		           brokerName,
#endif
				   rcString,
				   (char *)0);
		}
		else
		{
		    TRACE_MID(TRACE_INFO, ("Message successfully propagated"));
		}
	}
	else
	{
		if (rc == CCI_EXCEPTION)
		{
			cciRethrowLastException(0);
		}
		else
		{
			/* some other error, better log it */
			sprintf(msgBuffer, "Failed to propagate message (rc=%d)", rc);
			sprintf(rcString, "%d", rc);
			/* log error message */
			cciLog(&rc, 
				   CCI_LOG_ERROR, 
				   __FILE__, 
				   __LINE__, 
				   functionName, 
				   C.PLUGIN_MESSAGE_CATALOGUE, 
				   MJB_MESSAGE_NOT_PROPAGATED, 
				   msgBuffer,
#if !defined(WIN32)
		           brokerName,
#endif
				   rcString,
				   (char *)0);
		}
	}
    TRACE_RETURN(RSN_036, TT_SUCCESS);
}

/******************************************************************************/
/* copyMessage                                                                */
/******************************************************************************/
int copyMessage(IN CciMessage *newMsg, 
			    IN CciMessage *oldMsg)
{
	CciElement *newMsgRootP,
		       *oldMsgRootP;
	int rc;
    TRACE_ENTRY("copyMessage", ("newMsg=%.8x, oldMsg=%.8x",
		                        newMsg, oldMsg));

	/* get root element of both messages */
    newMsgRootP = cniRootElement(&rc, newMsg);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("cniRootElement for new message FAILED"));
		TRACE_RETURN(RSN_037, TT_FAILED_TO_GET_MESSAGE_ROOT);
	}
	TRACE_MID(TRACE_INFO, ("Addressed new message root element at %.8x", 
		                    newMsgRootP));

	/* Get the root element of old message */
    oldMsgRootP = cniRootElement(&rc, oldMsg);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("cniRootElement for old message FAILED"));
		TRACE_RETURN(RSN_038, TT_FAILED_TO_GET_MESSAGE_ROOT);
	}
	TRACE_MID(TRACE_INFO, ("Addressed old message root element at %.8x", 
		                    oldMsgRootP));

	/* copy original message to new message tree */
	cniCopyElementTree(&rc,
                       oldMsgRootP,
			           newMsgRootP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new message"));
		TRACE_RETURN(RSN_039, TT_FAILED_TO_COPY_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("Old message successfully copied to new"));

	TRACE_RETURN(RSN_040, TT_SUCCESS);
}

/******************************************************************************/
/* dispHex                                                                    */
/******************************************************************************/
void dispHex(IN char *inputString, 
			 IN int inputStringlength, 
			 OUT char *hexString)
{

	int i;
	char tempBuf[10];
    unsigned char *ch; 				        /* Current Buffer position */

    ch = (unsigned char *)inputString;

	/* start with null-terminated string */
	*hexString = '\0';

	for (i=0; i<inputStringlength; i++)
	{
		sprintf(tempBuf, "%0.2x", *ch++);
		strcat(hexString, tempBuf);
	}
    return;
}

/******************************************************************************/
/* swapMessageIdAndCorrelId                                                   */
/******************************************************************************/
int swapMessageIdAndCorrelId(INOUT CciMessage *message)
{
	int rc;
	char savedCorrelId[MQ_CORREL_ID_LENGTH],
	     savedMsgId[MQ_MSG_ID_LENGTH];
	CciElement *correlIdElementP,
		       *msgIdElementP;
    struct CciByteArray byteArrayDetails;
	CciSize numBytes;


    TRACE_ENTRY("swapMessageIdAndCorrelId", ("message=%.8x", message));

	/*
	 * locate and store the CorrelId 
	 */

	correlIdElementP = getMessageElement(message, CORREL_ID);
	if (!correlIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("CorrelId element not found in message"));
		TRACE_RETURN(RSN_043, TT_CORRELID_NOT_FOUND);
	}

	/* copy correl id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)savedCorrelId;
	byteArrayDetails.size = MQ_CORREL_ID_LENGTH;
	numBytes = cniElementByteArrayValue(&rc,
										correlIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Correl Id from message"));
		TRACE_RETURN(RSN_044, TT_CORRELID_NOT_FOUND);
	}

	if (traceEnabled)
	{
		char hexBuffer[2*MQ_CORREL_ID_LENGTH + 1];
		dispHex((char *)(byteArrayDetails.pointer),
				MQ_CORREL_ID_LENGTH,
				hexBuffer);
		TRACE_MID(TRACE_INFO, ("CorrelId value is %s", hexBuffer));
	}

	/*
	 * locate and store the MsgId 
	 */
	msgIdElementP = getMessageElement(message, MESSAGE_ID);
	if (!msgIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("MsgId element not found in message"));
		TRACE_RETURN(RSN_045, TT_MESSAGE_ID_NOT_FOUND);
	}

	/* copy message id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)savedMsgId;
	byteArrayDetails.size = MQ_MSG_ID_LENGTH;
	numBytes = cniElementByteArrayValue(&rc,
										msgIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Message Id from message"));
		TRACE_RETURN(RSN_046, TT_MESSAGE_ID_NOT_FOUND);
	}

	if (traceEnabled)
	{
		char hexBuffer[2*MQ_MSG_ID_LENGTH + 1];
		dispHex((char *)(byteArrayDetails.pointer),
				MQ_MSG_ID_LENGTH,
				hexBuffer);
		TRACE_MID(TRACE_INFO, ("MsgId value is %s", hexBuffer));
	}

	/*
	 * store MsgId into CorrelId 
	 */
	TRACE_MID(TRACE_INFO, ("Setting CorrelId to..."));
	TRACE_MIDHEX(TRACE_INFO, byteArrayDetails.pointer, MQ_MSG_ID_LENGTH);

	cniSetElementByteArrayValue(&rc,
								correlIdElementP,
								&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set CorrelId (cniSetElementByteArrayValue rc=%d)", rc));
		TRACE_RETURN(RSN_047, TT_FAILED_TO_SWAP_CORREL_AND_MSG_IDS);
	}
	TRACE_MID(TRACE_INFO, ("CorrelId successfully set to message id"));

	/*
	 * store CorrelId into MsgId 
	 */
	byteArrayDetails.pointer = (void *)savedCorrelId;
	byteArrayDetails.size = MQ_CORREL_ID_LENGTH;
	TRACE_MID(TRACE_INFO, ("Setting MsgId to..."));
	TRACE_MIDHEX(TRACE_INFO, byteArrayDetails.pointer, MQ_MSG_ID_LENGTH);

	cniSetElementByteArrayValue(&rc,
								msgIdElementP,
								&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set MsgId (cniSetElementByteArrayValue rc=%d)", rc));
		TRACE_RETURN(RSN_048, TT_FAILED_TO_SWAP_CORREL_AND_MSG_IDS);
	}
	TRACE_MID(TRACE_INFO, ("Message id successfully set to correl id"));

	TRACE_RETURN(RSN_049, TT_SUCCESS);
}

/******************************************************************************/
/* swapMessageReplyDetails                                                    */
/******************************************************************************/
int swapMessageReplyDetails(IN CciMessage *message, 
							IN CciChar *newReplyToQueue,
							IN CciChar *newReplyToQueueManager,
							OUT CciChar *oldReplyToQueue,
							IN int oldReplyToQueueLength,
							OUT CciChar *oldReplyToQueueManager,
							IN int oldReplyToQueueManagerLength)
{
	CciElement *msgRootP,
		       *rtqElementP,
		       *rtqmElementP;
	int rc;
	int valueLength;

    TRACE_ENTRY("swapMessageReplyDetails", ("message=%.8x, "
		                                    "newReplyToQueue='%S', "
		                                    "newReplyToQueueManager='%S', "
											"oldReplyToQueue=%.8x, "
											"oldReplyToQueueManager=%.8x, "
											"oldReplyToQueueLength=%d, "
											"oldReplyToQueueManagerLength=%d",
		                                    message, 
											newReplyToQueue, 
											newReplyToQueueManager,
											oldReplyToQueue,
											oldReplyToQueueManager,
											oldReplyToQueueLength,
											oldReplyToQueueManagerLength));


	if (oldReplyToQueue)
	{
		/* extract the current ReplyToQueue value */
		valueLength = oldReplyToQueueLength;
		rc = getMessageCharArrayValue(message,
									  REPLY_TO_QUEUE,
									  oldReplyToQueue,
									  &valueLength,
									  &rtqElementP);
		if (rc != TT_SUCCESS)
		{
			/* error */
			TRACE_MID(TRACE_ERROR, ("Failed to retrieve ReplyToQueue from message"));
			TRACE_RETURN(RSN_050, TT_FAILED_TO_GET_REPLY_TO_QUEUE_FROM_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("ReplyToQueue '%S' (%d bytes) successfully extracted "
							   "from old message", oldReplyToQueue, valueLength));
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Old ReplyToQueue name not saved"));
	}

	if (oldReplyToQueueManager)
	{
		/* extract the current ReplyToQueueManager value */
		valueLength = oldReplyToQueueManagerLength;
		rc = getMessageCharArrayValue(message,
									  REPLY_TO_QUEUE_MANAGER,
									  oldReplyToQueueManager,
									  &valueLength,
									  &rtqmElementP);
		if (rc != TT_SUCCESS)
		{
			/* error */
			TRACE_MID(TRACE_ERROR, ("Failed to retrieve ReplyToQueueManager from message"));
			TRACE_RETURN(RSN_051, TT_FAILED_TO_GET_REPLY_TO_QUEUE_MANAGER_FROM_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("ReplyToQueueManager '%S' (%d bytes) successfully extracted "
							   "from old message", oldReplyToQueueManager, valueLength));
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Old ReplyToQueueManager name not saved"));
	}

	/* Get the root element of message */
    msgRootP = cniRootElement(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("cniRootElement for message FAILED"));
		TRACE_RETURN(RSN_052, TT_FAILED_TO_GET_MESSAGE_ROOT);
	}
	TRACE_MID(TRACE_INFO, ("Addressed message root element at %.8x", 
		                    msgRootP));

	/* address the ReplyToQueue in message */
	rtqElementP = getMessageElement(message, REPLY_TO_QUEUE);
	rtqmElementP = getMessageElement(message, REPLY_TO_QUEUE_MANAGER);
    if (!rtqElementP || !rtqmElementP)
	{
		if (!rtqElementP)
		{
		    TRACE_RETURN(RSN_053, TT_FAILED_TO_GET_REPLY_TO_QUEUE_FROM_MESSAGE);
		}
		else
		{
		    TRACE_RETURN(RSN_054, TT_FAILED_TO_GET_REPLY_TO_QUEUE_MANAGER_FROM_MESSAGE);
		}
	}

	TRACE_MID(TRACE_INFO, ("Setting ReplyToQueue to '%S'",
						   newReplyToQueue));
	cniSetElementCharacterValue(&rc,
		                        rtqElementP,
								newReplyToQueue,
								MQ_Q_NAME_LENGTH);
	if (rc != CCI_SUCCESS)
	{
	    TRACE_MID(TRACE_ERROR, ("Failed to set ReplyToQueue "
			                    "(cniSetElementByteArrayValue rc=%d)", rc));
	    TRACE_RETURN(RSN_055, TT_FAILED_TO_SET_REPLY_TO_QUEUE_IN_MESSAGE);
	}
    TRACE_MID(TRACE_INFO, ("ReplyToQueue successfully set"));

	/* replace the ReplyToQueueManager with the supplied value */
	TRACE_MID(TRACE_INFO, ("Setting ReplyToQueueManager to '%S'",
						   newReplyToQueueManager));
	cniSetElementCharacterValue(&rc,
		                        rtqmElementP,
								newReplyToQueueManager,
								MQ_Q_NAME_LENGTH);

	if (rc != CCI_SUCCESS)
	{
	    TRACE_MID(TRACE_ERROR, ("Failed to set ReplyToQueueManager "
			                    "(cniSetElementByteArrayValue rc=%d)", rc));
	    TRACE_RETURN(RSN_056, TT_FAILED_TO_SET_REPLY_TO_QUEUE_MANAGER_IN_MESSAGE);
	}
    TRACE_MID(TRACE_INFO, ("ReplyToQueueManager successfully set"));

	/* return the new message to caller */
    TRACE_RETURN(RSN_057, TT_SUCCESS);
}

/******************************************************************************/
/* getMessageLength                                                           */
/******************************************************************************/
int getMessageLength(IN CciMessage *message,
					 OUT int *lengthP)
{
	CciSize bufferSize;
	int rc;
	CciMessageContext *msgContextP;
    CciMessage *newMsg = NULL;

    TRACE_ENTRY("getMessageLength", ("message=%.8x", message));

	bufferSize = cniBufferSize(&rc, message);

	/* if syntax tree has not been serialised, then serialise a COPY of the tree */
	if (!bufferSize)
	{
		/* make a copy of the message using current message as a template */
		msgContextP = cniGetMessageContext(&rc, message);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to get message context "
									"(cniGetMessageContext rc=%d)", rc));
	        TRACE_RETURN(RSN_058, TT_FAILED_TO_COPY_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

		newMsg = cniCreateMessage(&rc, msgContextP);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to create new message "
									"(cniCreateMessage rc=%d)", rc));
	        TRACE_RETURN(RSN_059, TT_FAILED_TO_COPY_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

		/* Copy old message into new */
		rc = copyMessage(newMsg, message);
		if (rc != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new message"));
	        TRACE_RETURN(RSN_060, TT_FAILED_TO_COPY_MESSAGE);
		}

		cniWriteBuffer(&rc, newMsg);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to serialize message to buffer "
									"(cniWriteBuffer rc=%d", rc));
	        TRACE_RETURN(RSN_061, TT_FAILED_TO_SERIALIZE_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("Message serialised to buffer"));

		bufferSize = cniBufferSize(&rc, newMsg);
	}

    TRACE_MID(TRACE_INFO, ("Message buffer: length=%d", bufferSize));

	/* free up message if a copy was made */
	if (newMsg)
	{
		TRACE_MID(TRACE_INFO, ("Message copy freed"));
		cniDeleteMessage(&rc, newMsg);
	}
	
	*lengthP = bufferSize;

    TRACE_RETURN(RSN_062, TT_SUCCESS);
}

/******************************************************************************/
/* traceMessageBuffer                                                         */
/******************************************************************************/
void traceMessageBuffer(IN CciMessage *message)
{
	char *bufferStart;
	CciSize bufferSize;
	int rc;
	CciMessageContext *msgContextP;
    CciMessage *newMsg = NULL;

    TRACE_ENTRY("traceMessageBuffer", ("message=%.8x", message));

	/* dump out new message buffer */
	bufferStart = (char *)cniBufferPointer(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("cniBufferPointer failed with rc=%d", rc));
		TRACE_RETURN_VOID(RSN_063);
	}
	bufferSize = cniBufferSize(&rc, message);

	/* if syntax tree has not been serialised, then serialise a COPY of the tree */
	if (!bufferSize)
	{
		/* make a copy of the message using current message as a template */
		msgContextP = cniGetMessageContext(&rc, message);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to get message context "
									"(cniGetMessageContext rc=%d)", rc));
			TRACE_RETURN_VOID(RSN_064);
		}
		TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

		newMsg = cniCreateMessage(&rc, msgContextP);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to create new message "
									"(cniCreateMessage rc=%d)", rc));
			TRACE_RETURN_VOID(RSN_065);
		}
		TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

		/* Copy old message into new */
		rc = copyMessage(newMsg, message);
		if (rc != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new message"));
			TRACE_RETURN_VOID(RSN_066);
		}

		cniWriteBuffer(&rc, newMsg);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to serialize message to buffer "
									"(cniWriteBuffer rc=%d", rc));
			TRACE_RETURN_VOID(RSN_067);
		}
		TRACE_MID(TRACE_INFO, ("Message serialised to buffer"));

		bufferStart = (char *)cniBufferPointer(&rc, newMsg);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("cniBufferPointer failed with rc=%d", rc));
			TRACE_RETURN_VOID(RSN_068);
		}
		bufferSize = cniBufferSize(&rc, newMsg);
	}

    TRACE_MID(TRACE_INFO, ("Message buffer: start=%.8x, length=%d, data is...",
			               bufferStart, bufferSize));
	TRACE_MIDHEX(TRACE_INFO, bufferStart, bufferSize);

	/* free up message if a copy was made */
	if (newMsg)
	{
		TRACE_MID(TRACE_INFO, ("Message copy freed"));
		cniDeleteMessage(&rc, newMsg);
	}

	TRACE_RETURN_VOID(RSN_069);
}

/******************************************************************************/
/* locateElementInMessage                                                     */
/******************************************************************************/
CciElement *locateElementInMessage(IN CciMessage *message,
								   IN CciChar *elementName,
								   IN bool_t createElement,
	 							   IN elementType_t *elementTypeP,
								   OUT bool_t *newElementCreatedP)
{
	CciElement *elementP;
	int rc;
	CciChar *pathP;
#if defined(POST_201_VERSION)
	CciMessage *environmentMessageP;
#endif
    TRACE_ENTRY("locateElementInMessage", ("message=%.8x, elementName=%S, createElement=%s, "
		                          "elementTypeP=%s",
		                           message, elementName,
								   createElement ? "TRUE" : "FALSE",
								   !elementTypeP ? "null" :
 	                           *elementTypeP == ELEMENT_TYPE_NAME ? "ELEMENT_TYPE_NAME" :
	                           *elementTypeP == ELEMENT_TYPE_VALUE ? "ELEMENT_TYPE_VALUE" :
	                           "ELEMENT_TYPE_NAME_VALUE"));

	/*
	 * Determine which tree the required element belongs to and address its root
	 */
	if (!UCncmp(elementName, U.ROOT, UClen(U.ROOT)))
	{
		/* 
		 * Assume elementName is of the form 'Root' or 'Root.<element path>' - address message
		 * root first and remove either 'Root' or 'Root.' prefix from elementName 
		 */
		elementP = cniRootElement(&rc, message);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("cniRootElement for message FAILED"));
			TRACE_RETURN_TYPE(RSN_069, CciElement *, NULL);
		}
		TRACE_MID(TRACE_INFO, ("Addressed  message root element at %.8x", 
								elementP));


		if (UClen(elementName) > UClen(U.ROOT))
		{
			/* strip off 'Root.' from elementName */
			pathP = UCchr(elementName, U.DOT_CHAR);
			pathP++;  /* move past the '.' */
			elementP = locateElementInTree(message,
										   elementP, 
										   pathP, 
										   createElement, 
										   elementTypeP,
										   TRUE, /* this is the firstElementUnderRoot */
										   newElementCreatedP);
			TRACE_RETURN_TYPE(RSN_069, CciElement *, elementP);
		}
	}

// cniGetEnvironmentMessage function only available from v2.0.2 onwards
#if defined(POST_201_VERSION)

	else
	if (!UCncmp(elementName, U.ENVIRONMENT, UClen(U.ENVIRONMENT)))
	{

		/* 
		 * Assume elementName is of the form 'Environment' or 'Environment.<element path>' 
		 * - address environment root first and remove either 'Environment' or 'Environment.' 
		 * prefix from elementName 
		 */
		environmentMessageP = cniGetEnvironmentMessage(&rc,
		                                               message);

		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("cniGetEnvironmentMessage for message FAILED"));
			TRACE_RETURN_TYPE(RSN_069, CciElement *, NULL);
		}
		TRACE_MID(TRACE_INFO, ("Addressed  Environment message tree at %.8x", 
								environmentMessageP));

		/* address the Environment root element */
		elementP = cniRootElement(&rc, environmentMessageP);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("cniRootElement for message FAILED"));
			TRACE_RETURN_TYPE(RSN_069, CciElement *, NULL);
		}
		TRACE_MID(TRACE_INFO, ("Addressed Environment message root element at %.8x", 
								elementP));

		if (UClen(elementName) > UClen(U.ENVIRONMENT))
		{
			/* strip off 'Environment.' from elementName */
			pathP = UCchr(elementName, U.DOT_CHAR);
			pathP++;  /* move past the '.' */
			elementP = locateElementInTree(message,
										   elementP, 
										   pathP, 
										   createElement, 
										   elementTypeP,
										   TRUE, /* this is the firstElementUnderRoot */
										   newElementCreatedP);
			TRACE_RETURN_TYPE(RSN_069, CciElement *, elementP);
		}
	}
#endif


	TRACE_RETURN_TYPE(RSN_069, CciElement *, elementP);
}

/******************************************************************************/
/* locateElementInTree                                                        */
/******************************************************************************/
CciElement *locateElementInTree(IN CciMessage *message,
								IN CciElement *currentElement,
								IN CciChar *elementNameAndPath,
								IN bool_t createElement,
								IN elementType_t *elementTypeP,
								IN bool_t firstElementUnderRoot,
								OUT bool_t *newElementCreatedP)
{
	CciChar nextElementName[MAX_STRING_SIZE + 1],
		    *dotP;
	CciElement *elementP;
	CciElementType eType;
	int rc;

    TRACE_ENTRY("locateElementInTree", ("message=%.8x, currentElement=%.8x, elementNameAndPath=%S, createElement=%s, "
		                          "elementType=%s, firstElementUnderRoot=%s",
		                           message, currentElement, elementNameAndPath,
								   createElement ? "TRUE" : "FALSE",
								   !elementTypeP ? "null" :
 	                           *elementTypeP == ELEMENT_TYPE_NAME ? "ELEMENT_TYPE_NAME" :
	                           *elementTypeP == ELEMENT_TYPE_VALUE ? "ELEMENT_TYPE_VALUE" :
	                           "ELEMENT_TYPE_NAME_VALUE",
								   firstElementUnderRoot ? "TRUE" : "FALSE"
								   ));

	/* assume element already exists */
	if (newElementCreatedP)
	{
		*newElementCreatedP = FALSE;
	}

	UCcpy(nextElementName, elementNameAndPath);

	/* strip off the next element name from the list and locate matching element in tree */
	dotP = UCchr(nextElementName, U.DOT_CHAR);

	if (dotP)
	{
		*dotP = 0;  /* replace dot with a NULL */
	}
	TRACE_MID(TRACE_INFO, ("Extracted element name is '%S'", nextElementName));

	/* find element as child of current element */
	TRACE_MID(TRACE_INFO, ("About to call cniSearchFirstChild with "
		                   "currentElement=%.8x, nextElementName=%S",
						   currentElement, nextElementName));
	elementP = cniSearchFirstChild(&rc,
		                           currentElement,
								   CCI_COMPARE_MODE_NAME,
							       CCI_ELEMENT_TYPE_NAME,
								   nextElementName);
	TRACE_MID(TRACE_INFO, ("Returned from cniSearchFirstChild"));
    if (!elementP)
	{
		/* element not found... */
		if (!createElement)
		{
			/* element not found in message - return NULL */
			TRACE_MID(TRACE_INFO, ("Requested element not found in message, returning NULL"));
			TRACE_RETURN_TYPE(RSN_069, CciElement *, NULL);
		}

		if (newElementCreatedP)
		{
			*newElementCreatedP = TRUE;
		}

		/* create new element with specified name */
		if (elementTypeP)
		{
			switch (*elementTypeP)
			{
				case ELEMENT_TYPE_NAME:
					eType = CCI_ELEMENT_TYPE_NAME;
					break;
				case ELEMENT_TYPE_VALUE:
					eType = CCI_ELEMENT_TYPE_VALUE;
					break;
				case ELEMENT_TYPE_NAME_VALUE:
					eType = CCI_ELEMENT_TYPE_NAME_VALUE;
					break;
				default:
					eType = CCI_ELEMENT_TYPE_UNKNOWN;
					break;
			}
		}
		else
		{
 			/* if this is the last element in the path then create a name-value element
			 * else just create a name element
			 */
		    if (!dotP) /* last name in path */
			{
				eType = CCI_ELEMENT_TYPE_NAME_VALUE;
			}
			else
			{
				eType = CCI_ELEMENT_TYPE_NAME;
			}
		}
		TRACE_MID(TRACE_INFO, ("Requested element not found in message, creating element"));

		if (!UCcmp(nextElementName, U.XML) && firstElementUnderRoot)
		{
			deleteMessageBody(message);
//			eType = 0x1000010;
		    /* create new body using the XML parser */ 
		    elementP = cniCreateElementAsLastChildUsingParser(&rc, 
                                                              currentElement,
                                                              U.XML);
			if (rc != CCI_SUCCESS) 
			{ 
				TRACE_MID(TRACE_ERROR, ("Failed to create new message XML body " 
										"(cniCreateElementAsLastChildUsingParser rc=%d)", rc)); 
				TRACE_RETURN_TYPE(RSN_070, CciElement *, NULL);
			} 
	        TRACE_MID(TRACE_INFO, ("New XML message body created at %.8x", elementP)); 
		}
		else
		if (!UCcmp(nextElementName, U.BLOB) && firstElementUnderRoot)
		{
			deleteMessageBody(message);
//			eType = 0x1000010;
		    /* create new body using the XML parser */ 
		    elementP = cniCreateElementAsLastChildUsingParser(&rc, 
                                                              currentElement,
                                                              U.NONE);
			if (rc != CCI_SUCCESS) 
			{ 
				TRACE_MID(TRACE_ERROR, ("Failed to create new message BLOB body " 
										"(cniCreateElementAsLastChildUsingParser rc=%d)", rc)); 
				TRACE_RETURN_TYPE(RSN_070, CciElement *, NULL);
			} 
	        TRACE_MID(TRACE_INFO, ("New BLOB message body created at %.8x", elementP)); 
		}
		else
		{
			elementP = createChildElement(currentElement,
										  nextElementName,
										  eType);
		}
		if (!elementP)
		{
			TRACE_RETURN_TYPE(RSN_069, CciElement *, NULL);
		}
	}
	else
	{
	    TRACE_MID(TRACE_INFO, ("Element '%S' located at %.8x", nextElementName, elementP));
	}

	/* if this is the last element in the path, then just return */
	if (!dotP)
	{
		/* no more elements in path, so return current element */
	    TRACE_MID(TRACE_INFO, ("End of path reached, returning element located at %.8x", 
			                    elementP));
		TRACE_RETURN_TYPE(RSN_069, CciElement *, elementP);
	}

	dotP++; /* move past the dot/null */

	/* call this function again, passing remainder of element name string */
	elementP = locateElementInTree(message,
		                           elementP, 
		                           dotP, 
								   createElement, 
								   elementTypeP, 
								   FALSE, /* not the firstElementUnderRoot */
								   newElementCreatedP);
    TRACE_MID(TRACE_INFO, ("Returning element located at %.8x", elementP)); 

	TRACE_RETURN_TYPE(RSN_069, CciElement *, elementP);
}

/******************************************************************************/
/* deleteMessageBody                                                          */
/******************************************************************************/
int deleteMessageBody(INOUT CciMessage *message)
{
	int rc;
	CciElement *mqmdElementP,
			   *siblingP;
    CciChar elementName[256];
    CciSize elementNameSize;

    TRACE_ENTRY("deleteMessageBody", ("message=%.8x", message));

	/* remove all sibling elements after MQMD */
	mqmdElementP = getMessageElement(message, MQMD_ELEMENT);

	if (!mqmdElementP)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to address MQMD element"));
		TRACE_RETURN(RSN_021, TT_FAILED_TO_GET_MESSAGE_MQMD);
	}
	TRACE_MID(TRACE_INFO, ("Addressed MQMD element at %.8x", mqmdElementP));

	/* delete all elements following the MQMD */
	siblingP = cniNextSibling(&rc, mqmdElementP);
	while(siblingP)
	{
		/* Get the element's name */
		elementNameSize = cniElementName(&rc, 
										 siblingP, 
										 (CciChar *)&elementName, 
										 sizeof(elementName));

		/* empty wc string seems to have a length of 1 */
		if (elementNameSize <= 1)
		{
			TRACE_MID(TRACE_INFO, ("Deleting element with no name"));
		}
		else
		{
		    TRACE_MID(TRACE_INFO, ("Deleting element with name %S", elementName));
		}

		cniDetach(&rc, siblingP);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to detach sibling element "
									"(cniDetach rc=%d)", rc));
			TRACE_RETURN(RSN_022, TT_FAILED_TO_DELETE_MESSAGE_BODY);
		}
		TRACE_MID(TRACE_INFO, ("Detached sibling element"));

		/* address the next sibling */
	    siblingP = cniNextSibling(&rc, mqmdElementP);
	}

	TRACE_RETURN(RSN_023, TT_SUCCESS);
}

/******************************************************************************/
/* serialiseSyntaxElements                                                    */
/******************************************************************************/
int serialiseSyntaxElements(IN CciElement *copyFromElement,
							IN bool_t storeAsXML,
							IN bool_t ignoreDTD,
							OUT char **outputBufferP,
							OUT int *lengthP)
{
	int retCode,
		bufferLength,
		depth,
		dataLength;
	char *buffer,
		 *bufferPos;

    TRACE_ENTRY("serialiseSyntaxElements", ("copyFromElement=%.8x, "
		                                    "storeAsXML=%s, ignoreDTD=%s, "
											"outputBufferP=%.8x", 
		                                     copyFromElement, 
											 storeAsXML ? "TRUE" : "FALSE", 
											 ignoreDTD ? "TRUE" : "FALSE", outputBufferP));

	/* allocate initial buffer, using 1K increments */
	bufferLength = ONE_KBYTE;
	buffer = (char *)malloc(bufferLength);
	TRACE_ALLOC("utilities.c-41", buffer);
	bufferPos = buffer;
	depth = 0;

	/* 
	 * Starting at the current element, walk through any of its descendent elements
	 * using a left, depth first approach.
	 * For each element encountered, sufficient details to enable it to be recreated.
	 */

	retCode = deflateAllSyntaxElements(copyFromElement,
		                               storeAsXML,
									   ignoreDTD,
		                               &buffer,
						   			   &bufferPos,
									   &bufferLength,
									   &depth);
	if (retCode != TT_SUCCESS)
	{
		free(buffer);
	    TRACE_FREE("utilities.c-41", buffer);
	    TRACE_RETURN(RSN_062, retCode);
	}

	/* calculate length of data in buffer */
    dataLength = (int)bufferPos - (int)buffer;
	if (dataLength < bufferLength)
	{
		buffer = (char *)realloc(buffer, dataLength);
		TRACE_ALLOC("utilities.c-41", buffer);
	}

    TRACE_MID(TRACE_INFO, ("Data length is %d", dataLength));
	*outputBufferP = buffer;
	*lengthP = dataLength;

    TRACE_RETURN(RSN_062, TT_SUCCESS);
}

/******************************************************************************/
/* deflateAllSyntaxElements                                                   */
/******************************************************************************/
int deflateAllSyntaxElements(IN CciElement *elementP,
							 IN bool_t storeAsXML,
							 IN bool_t ignoreDTD,
						     IN char **bufferP,
						     INOUT char **currentPosP,
						     INOUT int *bufferLengthP,
						     INOUT int *treeDepthP)
{
	CciElement *childP,
		       *siblingP;
    int retCode,
		rc,
		siblingCount;

    TRACE_ENTRY("deflateAllSyntaxElements", ("elementP=%.8x, storeAsXML=%s, "
		                                     "ignoreDTD=%s, "
		                                     "*bufferP=%.8x, *currentPosP=%.8x, "
		                                  "*bufferLengthP=%d, *treeDepthP=%d",
										  elementP, 
										  storeAsXML ? "TRUE" : "FALSE",
										  ignoreDTD ? "TRUE" : "FALSE",
										  *bufferP, *currentPosP, *bufferLengthP,
										  *treeDepthP));

	if (ignoreDTD && isElementDTD(elementP))
	{
        TRACE_MID(TRACE_INFO, ("Ignoring DTD element"));
	    TRACE_RETURN(RSN_062, TT_SUCCESS);
	}

	/* process current element data */
	retCode = deflateSyntaxElement(elementP,
		                           storeAsXML,
		                           bufferP,
						   		   currentPosP,
								   bufferLengthP,
								   treeDepthP);

	if (retCode != TT_SUCCESS)
	{
	    TRACE_RETURN(RSN_062, retCode);
	}

	/* process each of the children of this node */
    childP = cniFirstChild(&rc, elementP);
    if (!childP)
    { 
        TRACE_MID(TRACE_INFO, ("Node has no children"));
	    TRACE_RETURN(RSN_062, TT_SUCCESS);
    }
    TRACE_MID(TRACE_INFO, ("Got first child of node at %.8x", childP));

	*treeDepthP += 1;

    retCode = deflateAllSyntaxElements(childP, 
		                               storeAsXML,
		                               ignoreDTD,
						               bufferP,
									   currentPosP,
									   bufferLengthP,
									   treeDepthP);
	if (retCode != TT_SUCCESS)
	{
	    TRACE_RETURN(RSN_062, retCode);
	}

    siblingP = cniNextSibling(&rc, childP);
	siblingCount = 0;
	while (siblingP)
    {
        siblingCount++;
        TRACE_MID(TRACE_INFO, ("Processing sibling at %.8x, sibling count is %d", 
			                   siblingP, siblingCount));

		retCode = deflateAllSyntaxElements(siblingP,
										   storeAsXML,
										   ignoreDTD,
										   bufferP,
										   currentPosP,
										   bufferLengthP,
										   treeDepthP);
		if (retCode != TT_SUCCESS)
		{
			TRACE_RETURN(RSN_062, retCode);
		}
        siblingP = cniNextSibling(&rc, siblingP);
    }

	if (!siblingCount)
	{
        TRACE_MID(TRACE_INFO, ("No siblings present"));
	}

	*treeDepthP -= 1;

    TRACE_RETURN(RSN_062, TT_SUCCESS);
}

/******************************************************************************/
/* deflateSyntaxElement                                                       */
/******************************************************************************/
int deflateSyntaxElement(IN CciElement *elementP,
						 IN bool_t storeAsXML,
						 IN char **bufferP,
						 INOUT char **currentPosP,
						 INOUT int *bufferLengthP,
						 INOUT int *treeDepthP)
{
	elementHeader_t *headerP;
	CciElementType elementType;
	CciElement *childElementP;
	CciValueType valueType;
	CciSize nameSize;
	CciBool boolValue;
	CciInt intValue;
	CciReal realValue;
	struct CciTime timeValue;
	struct CciDate dateValue;
	struct CciTimestamp timeStampValue;
    int rc,
		bufferLength,
		currentDepth,
		sizeofHeader,
		length,
		numBytes,
		offsetForLengthField,
		spareBytesInBuffer,
		dataLength,
		retCode;
	char *buffer,
		 *bufferPos,
		 *oldBufferPos,
		 *dataP=NULL;
	CciChar name[MAX_STRING_SIZE + 1],
		    smallBuffer[1];
	bool_t elementHasAValue,
	       elementHasAName,
		   elementHasAChild,
		   lengthRequired;
    struct CciByteArray byteArrayDetails;
    struct CciBitArray bitArrayDetails;

    TRACE_ENTRY("deflateSyntaxElement", ("elementP=%.8x, storeAsXML=%s, "
		                                 "*bufferP=%.8x, *currentPosP=%.8x, "
		                                  "*bufferLengthP=%d, *treeDepthP=%d",
										  elementP, 
										  storeAsXML ? "TRUE" : "FALSE", 
										  *bufferP, *currentPosP, *bufferLengthP,
										  *treeDepthP));

	buffer = *bufferP;
	bufferLength = *bufferLengthP;
	bufferPos = *currentPosP;
	currentDepth = *treeDepthP;

	/* 
	 * For the current element, capture sufficient details to enable it to be recreated.
	 */

    elementHasAValue = FALSE;
	elementHasAName = FALSE;
	elementHasAChild = FALSE;

	/* determine the type of current element */
	elementType = cniElementType(&rc, 
		                         elementP);

	// check if it has a name
	nameSize = cniElementName(&rc, 
							  elementP, 
							  name, 
							  sizeof(name));

	// null name seems to have a length of 1 (presumably it includes the null) !!!
	if (nameSize > 1)
	{
		elementHasAName = TRUE;
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("element has no name"));
		elementHasAName = FALSE;
	}

	/* check if it contains any child elements */
	childElementP = cniFirstChild(&rc,
                                  elementP);
	if (childElementP)
	{
		elementHasAChild = TRUE;
	}

	/* check if element has a value */
	switch (elementType)
	{
	    case CCI_ELEMENT_TYPE_NAME:
            TRACE_MID(TRACE_INFO, ("element type is NAME"));
            break;
	    case CCI_ELEMENT_TYPE_VALUE:
            TRACE_MID(TRACE_INFO, ("element type is VALUE"));
            elementHasAValue = TRUE;
            break;
	    case CCI_ELEMENT_TYPE_NAME_VALUE:
            TRACE_MID(TRACE_INFO, ("element type is NAME_VALUE"));
            elementHasAValue = TRUE;
            break;
		/* 
		 * There are a range of elements which appear to fall outside those
		 * defined in BipCci.h.  For example, the DTD and its attributes have types
		 * which do not appear to be publicly defined.  Therefore, in the interest
		 * of expediency, we will assume that these types are indeed valid and will
		 * check them to see if they have a NAME and/or a VALUE.
		 */

		/* if this is the DTD tag, then it has a name but no value (even though a
		 * value can be extracted from it!).  Take for example the following message 
		 *
		 *	<?xml version="1.0" encoding="UTF-8"?><DATA>values</DATA>
		 *
		 *	This results in a message structure that looks like this...
		 *	MQMD as usual
		 *	(0x1000010)XML = (
		 *		(0x5000018)XML = (
		 *			(0x6000011)='1.0'
		 *			(0x6000012)='UTF-8'
		 *		)
		 *		(0x1000000)DATA = (
		 *			(0x2000000)='value'
		 *		)
		 *	)
		 *
		 * If the DTD XML element is interrogated, it will return a value of 
		 * '1.0UTF-8', i.e. a concatenation of its child element values.
		 * Therefore, we need a specific test for this particular element type.
		 */
	    case MJB_ELEMENT_TYPE_XML_BODY:
			TRACE_MID(TRACE_INFO, ("element type is XML body"));
			break;
	    case MJB_ELEMENT_TYPE_DTD_HEADER:
			TRACE_MID(TRACE_INFO, ("element type is DTD header"));
			break;
        default:
			TRACE_MID(TRACE_INFO, ("element type is 0x%.8x", elementType));

			/* determine what value type (if any) the element has */
			valueType = cniElementValueType(&rc, elementP);

			// We assume that all 'unknown' element types are related to attributes of
			// an XML DTD and will therefore have a value type of CCI_VALUE_TYPE_CHAR.
			// If the current element does not have this value type, then return an error 
			if (valueType != CCI_VALUE_TYPE_CHAR)
			{
				TRACE_MID(TRACE_ERROR, ("UNKNOWN element type with non CHAR value"));
				TRACE_RETURN(RSN_062, TT_UNKNOWN_ELEMENT_TYPE);
			}

			elementHasAValue = TRUE;
		    break;
    }

	/* allocate a new element header, allowing for optional Unicode name */
	sizeofHeader = sizeof(headerFixed_t);
	if (nameSize > 1)
	{
		sizeofHeader += ((UClen(name) + 1) * sizeof(CciChar));
	}
    TRACE_MID(TRACE_INFO, ("Allocating %d bytes for header%s", 
		                   sizeofHeader,
						   nameSize>1 ? " and element name" : ""));
	headerP = (elementHeader_t *)malloc(sizeofHeader);
	TRACE_ALLOC("utilities.c-42", headerP);
	memset(headerP, 0, sizeofHeader);
    
	headerP->fixedPart.headerLength = sizeofHeader;
	headerP->fixedPart.depth = currentDepth;
	headerP->fixedPart.elementType = elementType;

	if (elementHasAValue)
	{
		headerP->fixedPart.flagWord |= ELEMENT_HAS_A_VALUE;
		/* get element value type */
		valueType = cniElementValueType(&rc, elementP);
	    TRACE_MID(TRACE_INFO, ("valueType is %d", valueType));
		headerP->fixedPart.valueType = valueType;
	}

	if (elementHasAName)
	{
		TRACE_MID(TRACE_INFO, ("element name is '%S' (num chars = %d)", name, nameSize));
		headerP->fixedPart.flagWord |= ELEMENT_HAS_A_NAME;
		UCcpy(headerP->name, name);
		TRACE_MID(TRACE_INFO, ("element name copied into header"));
	}

	if (elementHasAChild)
	{
		TRACE_MID(TRACE_INFO, ("element has a child"));
		headerP->fixedPart.flagWord |= ELEMENT_HAS_A_CHILD;
	}

	/* write header into buffer */
	if (extendBufferIfRequired(headerP->fixedPart.headerLength, bufferP, 
		                       currentPosP, bufferLengthP))
	{
	    TRACE_MID(TRACE_INFO, ("Buffer was extended, resetting local pointers"));

		TRACE_MID(TRACE_INFO, ("*bufferP value is %.8x", *bufferP));
		buffer = *bufferP;
		TRACE_MID(TRACE_INFO, ("*bufferLengthP value is %d", *bufferLengthP));
		bufferLength = *bufferLengthP;
		TRACE_MID(TRACE_INFO, ("*currentPosP value is %.8x", *currentPosP));
		bufferPos = *currentPosP;
	}

	memcpy(bufferPos, headerP, headerP->fixedPart.headerLength);
	bufferPos += headerP->fixedPart.headerLength;
	*currentPosP = bufferPos;

	/* free up the header */
	free(headerP);
    TRACE_FREE("utilities.c-42", headerP);

	if (elementHasAValue)
	{
		if (storeAsXML)
		{
			/* extract the current value as string data */
			retCode = formatDataValueForXML(elementP,
									        NULL,   // no buffer specified, means get length only
									        0,      // no buffer length
									        &dataLength);

			/* increase buffer if required, allowing for terminating null */
			extendBufferIfRequired(dataLength + 1, 
				                   bufferP, 
								   currentPosP, 
								   bufferLengthP);

			/* go get the data */
			retCode = formatDataValueForXML(elementP,
									        *currentPosP,   // write data from current buffer position
									        *bufferLengthP, // use buffer length
									        &dataLength);
			if (retCode != TT_SUCCESS)
			{
				/* deal with errors */
				TRACE_MID(TRACE_ERROR, ("Error formatting data"));
				TRACE_RETURN(RSN_010, retCode);
			}

			/* add terminating NULL character */
			*(*currentPosP + dataLength) = '\0';

			/* move the buffer pointer past the data just copied in */
			*currentPosP += (dataLength+1);
			TRACE_MID(TRACE_INFO, ("bufferPos moved to %.8x", *currentPosP));

		    TRACE_RETURN(RSN_062, TT_SUCCESS);
		}

		/* 
		 * Allow for length field if data is variable length and not 
		 * CHAR type (which is null terminated)
		 */
		if (valueType == CCI_VALUE_TYPE_BYTEARRAY ||
			valueType == CCI_VALUE_TYPE_BITARRAY ||
			valueType == CCI_VALUE_TYPE_DECIMAL)
		{
			TRACE_MID(TRACE_INFO, ("Saving space in buffer for length value"));
			lengthRequired = TRUE;

			/* make sure theres enough space for the length field */
			if (extendBufferIfRequired(sizeof(int), bufferP, 
									   currentPosP, bufferLengthP))
			{
				TRACE_MID(TRACE_INFO, ("Buffer was extended for length value, resetting local pointers"));
				buffer = *bufferP;
				bufferLength = *bufferLengthP;
				bufferPos = *currentPosP;
			}

			/* remember where to save the length field */
			oldBufferPos = bufferPos;
			bufferPos += sizeof(int);
			*currentPosP = bufferPos;
            offsetForLengthField = oldBufferPos - buffer;
		}
		else
		{
			lengthRequired = FALSE;
		}

		/* determine how large the buffer needs to be */
		switch (valueType)
		{
			/****************************************************/ 
			/* The following are the variable length data types */
			/****************************************************/ 
			case CCI_VALUE_TYPE_BYTEARRAY:
			    byteArrayDetails.pointer = (void *)smallBuffer;
				byteArrayDetails.size = sizeof(smallBuffer);
				numBytes = cniElementByteArrayValue(&rc,
												    elementP,
												    &byteArrayDetails);
				break;

			case CCI_VALUE_TYPE_CHAR:
				numBytes = cniElementCharacterValue(&rc, 
										 		    elementP, 
												    smallBuffer,
						                            sizeof(smallBuffer));
				numBytes *= sizeof(CciChar);
					
				break;

			case CCI_VALUE_TYPE_BITARRAY:
			    bitArrayDetails.pointer = (void *)smallBuffer;
				bitArrayDetails.size = sizeof(smallBuffer);
				numBytes = cniElementBitArrayValue(&rc,
											       elementP,
											       &bitArrayDetails);
				break;

			case CCI_VALUE_TYPE_DECIMAL:
				numBytes = cniElementDecimalValue(&rc,
									              elementP,
									              smallBuffer,
												  sizeof(smallBuffer));
				break;

			/*************************************************/ 
			/* The following are the fixed length data types */
			/*************************************************/ 
			case CCI_VALUE_TYPE_BOOL:
				numBytes = sizeof(CciBool);
				break;

			case CCI_VALUE_TYPE_INTEGER:
				numBytes = sizeof(CciInt);
				break;

			case CCI_VALUE_TYPE_REAL:
				numBytes = sizeof(CciReal);
				break;

			case CCI_VALUE_TYPE_TIME:
			case CCI_VALUE_TYPE_GMTTIME:
				numBytes = sizeof(struct CciTime);
				break;

			case CCI_VALUE_TYPE_DATE:
				numBytes = sizeof(struct CciDate);
				break;

			case CCI_VALUE_TYPE_TIMESTAMP:
			case CCI_VALUE_TYPE_GMTTIMESTAMP:
				numBytes = sizeof(struct CciTimestamp);
				break;

			default:
				TRACE_MID(TRACE_INFO, ("element value type is (default) UNKNOWN"));
				break;
		}

		TRACE_MID(TRACE_INFO, ("numBytes set to %d", numBytes));

		/* make sure buffer is large enough to hold the data */
		if (extendBufferIfRequired(numBytes, bufferP, 
								   currentPosP, bufferLengthP))
		{
			TRACE_MID(TRACE_INFO, ("Buffer was extended for data, resetting local pointers"));
			buffer = *bufferP;
			bufferLength = *bufferLengthP;
			bufferPos = *currentPosP;

			if (lengthRequired)
			{
				/* re-calculate where to store the length value */
				oldBufferPos = buffer + offsetForLengthField;
			}
		}
		
		spareBytesInBuffer = bufferLength - ((int)bufferPos-(int)buffer);
		TRACE_MID(TRACE_INFO, ("Spare bytes in buffer is %d", spareBytesInBuffer));

		/* now copy in the element value */
		switch (valueType)
		{
			case CCI_VALUE_TYPE_BYTEARRAY:
				TRACE_MID(TRACE_INFO, ("element value type is BYTEARRAY"));
				byteArrayDetails.pointer = (void *)bufferPos;
				byteArrayDetails.size = spareBytesInBuffer;
				length = cniElementByteArrayValue(&rc,
												  elementP,
												  &byteArrayDetails);
				break;

			case CCI_VALUE_TYPE_CHAR:
				TRACE_MID(TRACE_INFO, ("element value type is CHAR"));
				length = cniElementCharacterValue(&rc, 
										 		  elementP, 
												 (CciChar*)bufferPos,
						                         spareBytesInBuffer);
				break;

			case CCI_VALUE_TYPE_BITARRAY:
				TRACE_MID(TRACE_INFO, ("element value type is BITARRAY"));
				bitArrayDetails.pointer = (void *)bufferPos;
				bitArrayDetails.size = spareBytesInBuffer;
				length = cniElementBitArrayValue(&rc,
												 elementP,
												 &bitArrayDetails);
				break;

			case CCI_VALUE_TYPE_DECIMAL:
				TRACE_MID(TRACE_INFO, ("element value type is DECIMAL"));
				length = cniElementDecimalValue(&rc,
									            elementP,
									            (CciChar *)bufferPos,
									            spareBytesInBuffer);
				break;

			/*************************************************/ 
			/* The following are the fixed length data types */
			/*************************************************/ 
			case CCI_VALUE_TYPE_BOOL:
				TRACE_MID(TRACE_INFO, ("element value type is BOOL"));
				boolValue = cniElementBooleanValue(&rc, elementP);
				dataP = (char *)&boolValue;
				break;

			case CCI_VALUE_TYPE_INTEGER:
				TRACE_MID(TRACE_INFO, ("element value type is INTEGER"));
				intValue = cniElementIntegerValue(&rc, elementP);
				dataP = (char *)&intValue;
				break;

			case CCI_VALUE_TYPE_REAL:
				TRACE_MID(TRACE_INFO, ("element value type is REAL"));
				realValue = cniElementRealValue(&rc, elementP);
				dataP = (char *)&realValue;
				break;

			case CCI_VALUE_TYPE_TIME:
				TRACE_MID(TRACE_INFO, ("element value type is TIME"));
				timeValue = cniElementTimeValue(&rc, elementP);
				dataP = (char *)&timeValue;
				break;

			case CCI_VALUE_TYPE_GMTTIME:
				TRACE_MID(TRACE_INFO, ("element value type is GMTTIME"));
				timeValue = cniElementGmtTimeValue(&rc, elementP);
				dataP = (char *)&timeValue;
				TRACE_MID(TRACE_INFO, ("Time is hh-mm-ss: %i-%i-%i",
					                   timeValue.hour,
									   timeValue.minute,
									   timeValue.second));
				break;

			case CCI_VALUE_TYPE_DATE:
				TRACE_MID(TRACE_INFO, ("element value type is DATE"));
				dateValue = cniElementDateValue(&rc, elementP);
				dataP = (char *)&dateValue;
				TRACE_MID(TRACE_INFO, ("Date is yy-mm-dd: %i-%i-%i",
					                   dateValue.year,
									   dateValue.month,
									   dateValue.day));
				break;

			case CCI_VALUE_TYPE_TIMESTAMP:
				TRACE_MID(TRACE_INFO, ("element value type is TIMESTAMP"));
				timeStampValue = cniElementTimestampValue(&rc, elementP);
				dataP = (char *)&timeStampValue;
				break;

			case CCI_VALUE_TYPE_GMTTIMESTAMP:
				TRACE_MID(TRACE_INFO, ("element value type is GMTTIMESTAMP"));
				timeStampValue = cniElementGmtTimestampValue(&rc, elementP);
				dataP = (char *)&timeStampValue;
				break;

			default:
				TRACE_MID(TRACE_INFO, ("element value type is (default) UNKNOWN"));
				break;
		}

		if (dataP)
		{
			TRACE_MID(TRACE_INFO, ("Copying %d bytes of fixed data type into buffer",
				                   numBytes));
			memcpy(*currentPosP, dataP, numBytes);
		}

		/* move the buffer pointer past the data just copied in */
		*currentPosP += numBytes;
		TRACE_MID(TRACE_INFO, ("bufferPos moved to %.8x", *currentPosP));

		/* copy in the length field if required */
		if (lengthRequired)
		{
		    /* copy the length value into buffer */
			memcpy(oldBufferPos, &numBytes, sizeof(int));
		}
	}

    TRACE_RETURN(RSN_062, TT_SUCCESS);
}


/******************************************************************************/
/* inflateSyntaxElement                                                       */
/******************************************************************************/
int inflateSyntaxElement(IN CciElement **elementPP,
						 IN char *buffer,
						 IN int bufferLength,
						 INOUT char **currentPosP,
						 INOUT int *treeDepthP,
						 IN bool_t firstElement,
						 IN bool_t includeFirstElement)
{
	elementHeader_t *headerP,
					workHeader,
					*bufHeaderP;
	bool_t elementHasAValue,
	       elementHasAName,
		   elementHasAChild,
		   correctTreeDepthFound;
	int depth,
		variableLength,
		rc,
		numCciChars;
	CciValueType valueType;
	void *dataP;
	CciElement *newElementP,
		       *currentElementP = *elementPP;
    struct CciByteArray byteArrayDetails;
    struct CciBitArray bitArrayDetails;
	bool_t unknownElementType = FALSE;

	CciInt workInt;
	CciReal workReal;
	CciBool workBool;
	struct CciTime workTime;
	struct CciDate workDate;
	struct CciTimestamp workTimestamp;

    TRACE_ENTRY("inflateSyntaxElement", ("*elementPP=%.8x, buffer=%.8x, bufferLength=%d, "
		                                 "*currentPosP=%.8x, *treeDepthP=%d, firstElement=%s, "
										 "includeFirstElement=%s",
										  *elementPP, buffer, bufferLength, *currentPosP,
										  *treeDepthP, 
										  firstElement ? "TRUE" : "FALSE",
										  includeFirstElement ? "TRUE" : "FALSE"));

    elementHasAValue = FALSE;
	elementHasAName = FALSE;
	elementHasAChild = FALSE;


	bufHeaderP = (elementHeader_t *)(*currentPosP);

	/* read the next header into local storage to ensure correct alignment */
	memcpy(&workHeader, *currentPosP, sizeof(elementHeader_t));

	/* address the next element header */
	headerP = &workHeader;

	switch (headerP->fixedPart.elementType)
	{
	    case CCI_ELEMENT_TYPE_NAME:
            TRACE_MID(TRACE_INFO, ("element type is NAME"));
            break;
	    case CCI_ELEMENT_TYPE_VALUE:
            TRACE_MID(TRACE_INFO, ("element type is VALUE"));
            break;
	    case CCI_ELEMENT_TYPE_NAME_VALUE:
            TRACE_MID(TRACE_INFO, ("element type is NAME_VALUE"));
            break;
	    case MJB_ELEMENT_TYPE_XML_BODY:
            TRACE_MID(TRACE_INFO, ("element type is XML body"));
            break;
		case MJB_ELEMENT_TYPE_DTD_HEADER:
            TRACE_MID(TRACE_INFO, ("element type is a DTD header"));
            break;
        default:
            TRACE_MID(TRACE_INFO, ("element type is (default) UNKNOWN (0x%.8x)",
				                    headerP->fixedPart.elementType));
			unknownElementType = TRUE;
		    break;
    }

    depth = headerP->fixedPart.depth;
	valueType = headerP->fixedPart.valueType;
    TRACE_MID(TRACE_INFO, ("depth=%d, valueType=%d", depth, valueType));

	if (headerP->fixedPart.flagWord & ELEMENT_HAS_A_NAME)
	{
		elementHasAName = TRUE;
	    TRACE_MID(TRACE_INFO, ("element name is %S", bufHeaderP->name));
	}
	else
	{
	    TRACE_MID(TRACE_INFO, ("element has no name"));
	}

	if (headerP->fixedPart.flagWord & ELEMENT_HAS_A_VALUE)
	{
		elementHasAValue = TRUE;
	}

	if (headerP->fixedPart.flagWord & ELEMENT_HAS_A_CHILD)
	{
		elementHasAChild = TRUE;
	}

	/* if this is not the first element, then create a new element */
	if (!firstElement || (firstElement && includeFirstElement))
	{
		correctTreeDepthFound = FALSE;
		while (!correctTreeDepthFound)
		{
			TRACE_MID(TRACE_INFO, ("depth=%d, *treeDepthP=%d", depth, *treeDepthP));
			if (depth == *treeDepthP)
			{
				/* create a sibling element - same depth */
				newElementP = createSiblingElement(*elementPP,
					              elementHasAName ? bufHeaderP->name : U.NULL_STRING,
							      headerP->fixedPart.elementType);
			}
			else
			if (depth > *treeDepthP)
			{
				/* create a new child element as last child - build tree from left to right */
				newElementP = createChildElement(*elementPP,
					              elementHasAName ? bufHeaderP->name : U.NULL_STRING,
								  headerP->fixedPart.elementType);

				/* increment the depth if element has children */
				if (elementHasAChild)
				{
					(*treeDepthP)++;
					TRACE_MID(TRACE_INFO, ("treeDepth increased to %d", *treeDepthP));
				}
				else
				{
					TRACE_MID(TRACE_INFO, ("Element has no children - depth not increased"));
				}
			}
			else
			{
				/* need to rewind back up the tree until depth == treeDepth */
				while (depth < *treeDepthP)
				{
					newElementP = cniParent(&rc,
											currentElementP);
					if (!newElementP)
					{
						/* deal with error here... */
						TRACE_RETURN(RSN_069, TT_FAILED_TO_GET_PARENT_ELEMENT);
					}
					(*treeDepthP)--;
					currentElementP = newElementP;
					*elementPP = currentElementP;
					TRACE_MID(TRACE_INFO, ("Parent element is %.8x, tree level "
						                   "decremented to %d", 
										   currentElementP, *treeDepthP));
				}
				continue;
			}
			correctTreeDepthFound = TRUE;
		}

		if (!newElementP)
		{
			TRACE_RETURN(RSN_069, TT_FAILED_TO_CREATE_NEW_ELEMENT);
		}

		/* address the new element */
		currentElementP = newElementP;

		/* update global element pointer if this is not a VALUE type element or unknown element */
		if (elementHasAChild)
		{
			*elementPP = newElementP;
			TRACE_MID(TRACE_INFO, ("*elementPP set to %.8x", newElementP));
		}
		else
		{
			TRACE_MID(TRACE_INFO, ("Element has no children - global pointer not reset"));
		}
	}

    /* move past the header */
	(*currentPosP) += headerP->fixedPart.headerLength;
    TRACE_MID(TRACE_INFO, ("Buffer pos moved on %d bytes to %.8x",
		                    headerP->fixedPart.headerLength,
							*currentPosP));

	if (elementHasAValue)
	{
	    TRACE_MID(TRACE_INFO, ("Element has a value"));
		/* allow for length value if this is a variable length data item */
		if (valueType == CCI_VALUE_TYPE_BYTEARRAY ||
			valueType == CCI_VALUE_TYPE_BITARRAY ||
			valueType == CCI_VALUE_TYPE_DECIMAL)
		{
			memcpy(&variableLength, *currentPosP, sizeof(int));
		    TRACE_MID(TRACE_INFO, ("Variable length value set to %d", variableLength));
			*currentPosP += sizeof(int);
			TRACE_MID(TRACE_INFO, ("Buffer pos moved on %d bytes to %.8x",
									sizeof(int),
									*currentPosP));
		}

		/* address the start of the data value in buffer */
		dataP = (void *)(*currentPosP);

		/* apply the data to the element */
		switch (valueType)
		{
			/****************************************************/ 
			/* The following are the variable length data types */
			/****************************************************/ 
			case CCI_VALUE_TYPE_BYTEARRAY:
				TRACE_MID(TRACE_INFO, ("element value type is BYTEARRAY"));
			    byteArrayDetails.pointer = dataP;
				byteArrayDetails.size = variableLength;
				cniSetElementByteArrayValue(&rc,
											currentElementP,
											&byteArrayDetails);
				*currentPosP += variableLength;
				break;

			case CCI_VALUE_TYPE_CHAR:
				TRACE_MID(TRACE_INFO, ("element value type is CHAR"));
				numCciChars = UClen((CciChar *)dataP);
				cniSetElementCharacterValue(&rc, 
										 	currentElementP, 
											dataP,
											numCciChars);
				*currentPosP += (numCciChars+1) * sizeof(CciChar);
				break;

			case CCI_VALUE_TYPE_BITARRAY:
				TRACE_MID(TRACE_INFO, ("element value type is BITARRAY"));
			    bitArrayDetails.pointer = dataP;
				bitArrayDetails.size = variableLength;
				cniSetElementBitArrayValue(&rc,
										   currentElementP,
										   &bitArrayDetails);
				*currentPosP += variableLength;
				break;

			case CCI_VALUE_TYPE_DECIMAL:
				TRACE_MID(TRACE_INFO, ("element value type is DECIMAL"));
				cniSetElementDecimalValue(&rc,
						                  currentElementP,
										  (CciChar *)dataP);
				*currentPosP += variableLength;
				break;

			/*************************************************/ 
			/* The following are the fixed length data types */
			/*************************************************/ 
			case CCI_VALUE_TYPE_BOOL:
				TRACE_MID(TRACE_INFO, ("element value type is BOOL"));
				/* Copy from buffer to local bool to ensure correct alignment */
				memcpy(&workBool, dataP, sizeof(CciBool));
				cniSetElementBooleanValue(&rc, 
				                          currentElementP,
										  workBool);
				*currentPosP += sizeof(CciBool);
				break;

			case CCI_VALUE_TYPE_INTEGER:
				TRACE_MID(TRACE_INFO, ("element value type is INTEGER"));
				/* Copy from buffer to local int to ensure correct alignment */
				memcpy(&workInt, dataP, sizeof(CciInt));
				cniSetElementIntegerValue(&rc, 
				                          currentElementP,
										  workInt);
				*currentPosP += sizeof(CciInt);
				break;

			case CCI_VALUE_TYPE_REAL:
				TRACE_MID(TRACE_INFO, ("element value type is REAL"));
				/* Copy from buffer to local real to ensure correct alignment */
				memcpy(&workReal, dataP, sizeof(CciReal));
				cniSetElementRealValue(&rc, 
				                       currentElementP,
									   workReal);
				*currentPosP += sizeof(CciReal);
				break;

			case CCI_VALUE_TYPE_TIME:
				TRACE_MID(TRACE_INFO, ("element value type is TIME"));
				/* Copy from buffer to local time struct to ensure correct alignment */
				memcpy(&workTime, dataP, sizeof(struct CciTime));
				cniSetElementTimeValue(&rc, 
				                       currentElementP,
									   &workTime);
				*currentPosP += sizeof(struct CciTime);
				break;

			case CCI_VALUE_TYPE_GMTTIME:
				TRACE_MID(TRACE_INFO, ("element value type is GMTTIME"));
				/* Copy from buffer to local time struct to ensure correct alignment */
				memcpy(&workTime, dataP, sizeof(struct CciTime));
				cniSetElementGmtTimeValue(&rc, 
				                          currentElementP,
									      &workTime);
				*currentPosP += sizeof(struct CciTime);
				break;

			case CCI_VALUE_TYPE_DATE:
				TRACE_MID(TRACE_INFO, ("element value type is DATE"));
				/* Copy from buffer to local date struct to ensure correct alignment */
				memcpy(&workDate, dataP, sizeof(struct CciDate));
				cniSetElementDateValue(&rc, 
				                       currentElementP,
								       &workDate);
				*currentPosP += sizeof(struct CciDate);
				break;

			case CCI_VALUE_TYPE_TIMESTAMP:
				TRACE_MID(TRACE_INFO, ("element value type is TIMESTAMP"));
				/* Copy from buffer to local timestamp struct to ensure correct alignment */
				memcpy(&workTimestamp, dataP, sizeof(struct CciTimestamp));
				cniSetElementTimestampValue(&rc, 
				                            currentElementP,
								            &workTimestamp);
				*currentPosP += sizeof(struct CciTimestamp);
				break;

			case CCI_VALUE_TYPE_GMTTIMESTAMP:
				TRACE_MID(TRACE_INFO, ("element value type is GMTTIMESTAMP"));
				/* Copy from buffer to local timestamp struct to ensure correct alignment */
				memcpy(&workTimestamp, dataP, sizeof(struct CciTimestamp));
				cniSetElementGmtTimestampValue(&rc, 
				                               currentElementP,
								               &workTimestamp);
				*currentPosP += sizeof(struct CciTimestamp);
				break;

			default:
				TRACE_MID(TRACE_INFO, ("element value type is (default) UNKNOWN"));
				break;
		}
	}
	else
	{
	    TRACE_MID(TRACE_INFO, ("Element has no value"));
	}

    TRACE_RETURN(RSN_062, TT_SUCCESS);
}


/******************************************************************************/
/* isElementDTD                                                               */
/******************************************************************************/
bool_t isElementDTD(IN CciElement *elementP)
{
	CciElementType elementType;
	int rc;

    TRACE_ENTRY("isElementDTD", ("elementP=%.8x",
		                          elementP));

	/* determine the type of current element */
	elementType = cniElementType(&rc, 
		                         elementP);

	if (elementType == MJB_ELEMENT_TYPE_DTD_HEADER)
	{
		TRACE_MID(TRACE_INFO, ("element type is DTD header"));
	    TRACE_RETURN_TYPE(RSN_062, bool_t, TRUE);
	}

    TRACE_RETURN_TYPE(RSN_062, bool_t, FALSE);
}


/******************************************************************************/
/* retrieveElementValue                                                       */
/******************************************************************************/
int retrieveElementValue(IN CciElement *elementP,
						 IN bool_t lookForChildValueElement,
						 OUT char **bufferP,
						 OUT int *bufferLengthP)
{
	CciElementType elementType,
		           childElementType;
	CciElement *childElementP;
	CciSize nameSize;
    int rc,
		retCode,
		dataLength;
	CciChar name[MAX_STRING_SIZE + 1];

    TRACE_ENTRY("retrieveElementValue", ("elementP=%.8x, lookForChildValueElement=%s", 
		                                  elementP,
                                          lookForChildValueElement ? "TRUE" : "FALSE"));
	/* 
	 * Extract the value of the specified element
	 */

    /* Get the element's name */
	nameSize = cniElementName(&rc, 
		                      elementP, 
							  name, 
							  sizeof(name));

    TRACE_MID(TRACE_INFO, ("element name is '%S' (num chars = %d)", name, nameSize));

	/* determine the type of current element */
	elementType = cniElementType(&rc, 
		                         elementP);
	switch (elementType)
	{
	    case CCI_ELEMENT_TYPE_NAME:
            TRACE_MID(TRACE_INFO, ("element type is NAME"));
            TRACE_MID(TRACE_ERROR, ("Element has no value"));

			/* check for a child element of type VALUE if allowed */
			if (lookForChildValueElement)
			{
				/* address the first (and assume only) child element */
			    childElementP = cniFirstChild(&rc, elementP);
				if (!childElementP)
				{ 
					TRACE_MID(TRACE_INFO, ("Node has no VALUE type child node"));
				    TRACE_RETURN(RSN_062, TT_ELEMENT_HAS_NO_VALUE);
				}
				TRACE_MID(TRACE_INFO, ("Addressed child node at %.8x", childElementP));

				/* check if it is a VALUE type node */
				childElementType = cniElementType(&rc, 
											      childElementP);
				if (childElementType != CCI_ELEMENT_TYPE_VALUE)
				{
					TRACE_MID(TRACE_INFO, ("Child node is not a VALUE type"));
				    TRACE_RETURN(RSN_062, TT_ELEMENT_HAS_NO_VALUE);
				}

				/* go get the child element value */
				retCode = retrieveElementValue(childElementP,
					                           FALSE,   // not interested in further child nodes
											   bufferP,
											   bufferLengthP);

			    TRACE_RETURN(RSN_062, retCode);
			}
		    TRACE_RETURN(RSN_062, TT_ELEMENT_HAS_NO_VALUE);
            break;
	    case CCI_ELEMENT_TYPE_VALUE:
            TRACE_MID(TRACE_INFO, ("element type is VALUE"));
            break;
	    case CCI_ELEMENT_TYPE_NAME_VALUE:
            TRACE_MID(TRACE_INFO, ("element type is NAME_VALUE"));
            break;
        default:
            TRACE_MID(TRACE_ERROR, ("element type is (default) UNKNOWN"));
		    TRACE_RETURN(RSN_062, TT_UNKNOWN_ELEMENT_TYPE);
		    break;
    }

	/* find out the size of the data */
	retCode = formatDataValueForXML(elementP,
		                            NULL,   // no buffer specified, means get length only
						            0,      // no buffer length
						            &dataLength);

	/* allocate buffer to return to caller */
	*bufferP = (char *)malloc(dataLength+1);
	TRACE_ALLOC("utilities.c-43", *bufferP);
	*bufferLengthP = dataLength+1;
	memset(*bufferP, 0, *bufferLengthP);
	TRACE_MID(TRACE_INFO, ("Output buffer of %d bytes allocated at %.8x", 
		                   *bufferLengthP,
						   *bufferP));

	/* copy extracted data into users buffer */
	retCode = formatDataValueForXML(elementP,
								    *bufferP, 
								    *bufferLengthP,
								    &dataLength);

	TRACE_MID(TRACE_INFO, ("Returning %d bytes of data", dataLength));

    TRACE_RETURN(RSN_062, TT_SUCCESS);
}


/******************************************************************************/
/* extendBufferIfRequired                                                     */
/******************************************************************************/
bool_t extendBufferIfRequired(IN int storageRequired,
							  INOUT char **bufferP,
						      INOUT char **currentPosP,
						      INOUT int *bufferLengthP)
{
	int numRemainingBytes,
		currentOffset,
		numKbytes;
    TRACE_ENTRY("extendBufferIfRequired", ("storageRequired=%d, *bufferP=%.8x, "
		                                   "*currentPosP=%.8x, *bufferLengthP=%d",
										  storageRequired, *bufferP, 
										  *currentPosP, *bufferLengthP));

	numRemainingBytes = *bufferLengthP - ((int)(*currentPosP) - (int)(*bufferP));
    TRACE_MID(TRACE_INFO, ("Buffer contains %d unused bytes", numRemainingBytes));
	if (numRemainingBytes < storageRequired)
	{
		currentOffset = (int)(*currentPosP) - (int)(*bufferP);
	    TRACE_MID(TRACE_INFO, ("Extending buffer storage, currentOffset is %d", currentOffset));
		if (storageRequired - numRemainingBytes > ONE_KBYTE)
		{
		    TRACE_MID(TRACE_INFO, ("More than ONE_KBYTE"));
			numKbytes = (storageRequired-numRemainingBytes) / ONE_KBYTE + 1;
		    TRACE_MID(TRACE_INFO, ("numKbytes is %d", numKbytes));
			*bufferP = (char *)realloc(*bufferP, *bufferLengthP + numKbytes * ONE_KBYTE);
			*bufferLengthP += numKbytes * ONE_KBYTE;
		}
		else
		{
		    TRACE_MID(TRACE_INFO, ("Just ONE_KBYTE"));
			*bufferP = (char *)realloc(*bufferP, *bufferLengthP + ONE_KBYTE);
			*bufferLengthP += ONE_KBYTE;
		}

		*currentPosP = *bufferP + currentOffset;

	    TRACE_RETURN_TYPE(RSN_062, bool_t, TRUE);
	}

    TRACE_RETURN_TYPE(RSN_062, bool_t, FALSE);
}

/******************************************************************************/
/* createChildElement                                                         */
/******************************************************************************/
CciElement *createChildElement(IN CciElement *parentElementP,
						       IN CciChar *elementName,
							   IN CciElementType elementType)
{
	int rc;
	CciElement *childElementP;
	bool_t unknownElementType = FALSE;
	char tempBuffer[MAX_STRING_SIZE+1];

    TRACE_ENTRY("createChildElement", ("parentElementP=%.8x, elementName=%S, "
		                               "elementType=%s",
									   parentElementP, 
									   elementName ? elementName : U.NULL_LITERAL,
 	                           elementType == CCI_ELEMENT_TYPE_NAME ? "CCI_ELEMENT_TYPE_NAME" :
	                           elementType == CCI_ELEMENT_TYPE_VALUE ? "CCI_ELEMENT_TYPE_VALUE" :
	                           elementType == CCI_ELEMENT_TYPE_NAME_VALUE ? "CCI_ELEMENT_TYPE_NAME_VALUE" :
	                           "CCI_ELEMENT_TYPE_UNKNOWN"));

	/* check for unknown element type */
    if (elementType != CCI_ELEMENT_TYPE_NAME &&
        elementType != CCI_ELEMENT_TYPE_VALUE &&
        elementType != CCI_ELEMENT_TYPE_NAME_VALUE)
	{
		unknownElementType = TRUE;
	}

	/* create new child element */
	childElementP = cniCreateElementAsLastChild(&rc,
	                                            parentElementP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create child element "
			                    "(cniCreateElementAsLastChild rc=%d)", rc));
	    TRACE_RETURN_TYPE(RSN_062, CciElement *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("Child element created at %.8x", childElementP));

	if (elementType != CCI_ELEMENT_TYPE_VALUE && elementName)
	{
		cniSetElementName(&rc,
						  childElementP,
						  elementName);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to set child element name "
									"(cniSetElementName rc=%d)", rc));
			TRACE_RETURN_TYPE(RSN_062, CciElement *, NULL);
		}
		TRACE_MID(TRACE_INFO, ("Child element name set to %S", elementName));
	}

	cniSetElementType(&rc,
		              childElementP,
					  elementType);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set child element type "
			                    "(cniSetElementType rc=%d)", rc));
	    TRACE_RETURN_TYPE(RSN_062, CciElement *, NULL);
	}

	if (unknownElementType)
	{
		sprintf(tempBuffer, " (0x%.8x)", elementType);
	}

	TRACE_MID(TRACE_INFO, ("Child element type set to %s",
 	                       elementType == CCI_ELEMENT_TYPE_NAME ? "CCI_ELEMENT_TYPE_NAME" :
	                       elementType == CCI_ELEMENT_TYPE_VALUE ? "CCI_ELEMENT_TYPE_VALUE" :
	                       elementType == CCI_ELEMENT_TYPE_NAME_VALUE ? "CCI_ELEMENT_TYPE_NAME_VALUE" :
						   unknownElementType ? tempBuffer : "CCI_ELEMENT_TYPE_UNKNOWN"));

    TRACE_RETURN_TYPE(RSN_062, CciElement *, childElementP);
}

/******************************************************************************/
/* createSiblingElement                                                       */
/******************************************************************************/
CciElement *createSiblingElement(IN CciElement *parentElementP,
						         IN CciChar *elementName,
							     IN CciElementType elementType)
{
	int rc;
	CciElement *siblingElementP;

    TRACE_ENTRY("createSiblingElement", ("parentElementP=%.8x, elementName=%S, "
		                               "elementType=%s",
		                                parentElementP, elementName,
						elementType==CCI_ELEMENT_TYPE_UNKNOWN ? "CCI_ELEMENT_TYPE_UNKNOWN" :
						elementType==CCI_ELEMENT_TYPE_NAME ? "CCI_ELEMENT_TYPE_NAME" :
						elementType==CCI_ELEMENT_TYPE_VALUE ? "CCI_ELEMENT_TYPE_VALUE" :
						elementType==CCI_ELEMENT_TYPE_NAME_VALUE ? "CCI_ELEMENT_TYPE_NAME_VALUE" :
	                                    "unknown"));

	/* create new child element */
	siblingElementP = cniCreateElementAfter(&rc,
	                                      parentElementP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create sibling element "
			                    "(cniCreateElementAfter rc=%d)", rc));
	    TRACE_RETURN_TYPE(RSN_062, CciElement *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("Sibling element created at %.8x", siblingElementP));
	cniSetElementName(&rc,
		              siblingElementP,
					  elementName);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set child element name "
			                    "(cniSetElementName rc=%d)", rc));
	    TRACE_RETURN_TYPE(RSN_062, CciElement *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("Child element name set to %S", elementName));
	cniSetElementType(&rc,
		              siblingElementP,
					  elementType);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set child element type "
			                    "(cniSetElementType rc=%d)", rc));
	    TRACE_RETURN_TYPE(RSN_062, CciElement *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("Child element type set"));

    TRACE_RETURN_TYPE(RSN_062, CciElement *, siblingElementP);
}


/******************************************************************************/
/* getSharedSegmentSize                                                       */
/******************************************************************************/
int getSharedSegmentSize(IN char *environmentVariableName,
						 IN int defaultSize)
{
	int requestedSize;
	char *smControlString;

    TRACE_ENTRY("getSharedSegmentSize", ("environmentVariableName=%s, defaultSize=%dK",
		                                  environmentVariableName, defaultSize));

// The next line was introduced to force the shared memory segment down to 2K in 
// order to force an SOS error for testing purposes
//    TRACE_RETURN_TYPE(RSN_001, int, 20*1024);
	

	smControlString = getenv(environmentVariableName); 

	if (!smControlString)
	{
		TRACE_MID(TRACE_INFO, ("Environment variable %s not specified - default value of "
			                   "%dK will be used", environmentVariableName, defaultSize));
	    TRACE_RETURN_TYPE(RSN_001, int, defaultSize*1024);
	}

	/*
	 * parse the shared memory size string                     
	 */
	requestedSize = atoi(smControlString);
	TRACE_MID(TRACE_INFO, ("Requested size is %dK", requestedSize));

	if (requestedSize < defaultSize)
	{
		TRACE_MID(TRACE_INFO, ("Requested size %dK is less than minimum size %dK - "
			                   "resetting to minimum size", 
							   requestedSize, defaultSize));
		requestedSize = defaultSize;
	}

	TRACE_RETURN_TYPE(RSN_001, int, requestedSize * 1024);
}


/******************************************************************************/
/* buildXMLFromSerialisedTree                                                 */
/******************************************************************************/
int buildXMLFromSerialisedTree(IN void *serialisedData,
							   IN int serialisedDataLength,
							   OUT char **XMLP,
							   OUT int *XMLLengthP)
{
	char *currentPos = NULL,
		 *dataBuffer = NULL,
		 *endPos = NULL,
		 *XMLCurrentPos;
	int retCode,
		dataBufferLength,
		depth = 0;

    TRACE_ENTRY("buildXMLFromSerialisedTree", ("serialisedData=%.8x, serialisedDataLength=%d",
		                                       serialisedData,
									    	   serialisedDataLength));

	/* create initial XML message buffer */
	*XMLLengthP = ONE_KBYTE;
	*XMLP = (char *)malloc(*XMLLengthP);
	TRACE_ALLOC("utilities.c-44", *XMLP);
	XMLCurrentPos = *XMLP;

	/* run through the serialised data and build it into an XML message */
	dataBuffer = serialisedData;
	currentPos = dataBuffer;
	dataBufferLength = serialisedDataLength;
	endPos = dataBuffer + dataBufferLength;

	retCode = convertSyntaxElementsToXML(XMLP,
			                             XMLLengthP,
									 	 &XMLCurrentPos,
									     dataBuffer,
									     dataBufferLength,
									     &currentPos,
										 endPos,
									     &depth);

	/* calculate the size of the XML message */
	TRACE_MID(TRACE_INFO, ("XMLCurrentPos is %.8x, *XMLP is %.8x", XMLCurrentPos, *XMLP));
	*XMLLengthP = (unsigned long)XMLCurrentPos - (unsigned long)(*XMLP);
	TRACE_MID(TRACE_INFO, ("Calculated XML message length is %d", *XMLLengthP));

    TRACE_RETURN(RSN_007, retCode);
}


/******************************************************************************/
/* processXMLTagEntry                                                         */
/******************************************************************************/
int processXMLTagEntry(IN char **XMLP,
				       IN int *XMLLengthP,
				       IN char **currentXMLP,
				       tagEntry_t *tagEntryP)
{
    TRACE_ENTRY("processXMLTagEntry", ("*XMLP=%.8x, *XMLLengthP=%d, "
		                               "*currentXMLP=%.8x, tagEntryP=%.8x",
								       *XMLP, *XMLLengthP, *currentXMLP,
								       tagEntryP));

	TRACE_MID(TRACE_INFO, ("Tag name is %s", tagEntryP->tagName));
	/* create closing tag for element */
	if (extendBufferIfRequired(strlen(tagEntryP->tagName) + 3,
							   XMLP,
							   currentXMLP,
							   XMLLengthP))
	{
		TRACE_MID(TRACE_INFO, ("Buffer was extended"));
		TRACE_MID(TRACE_INFO, ("*XMLP value is %.8x", *XMLP));
		TRACE_MID(TRACE_INFO, ("*XMLLengthP value is %d", *XMLLengthP));
		TRACE_MID(TRACE_INFO, ("*currentXMLP value is %.8x", *currentXMLP));
	}

	/* construct the closing tag in the XML buffer */
	memcpy(*currentXMLP, "</", 2);
	(*currentXMLP) += 2;

	memcpy(*currentXMLP, tagEntryP->tagName, strlen(tagEntryP->tagName));
	(*currentXMLP) += strlen(tagEntryP->tagName);

	**currentXMLP = '>';
	(*currentXMLP)++;

    TRACE_RETURN(RSN_007, TT_SUCCESS);
}

/******************************************************************************/
/* processXMLTagList                                                          */
/******************************************************************************/
int processXMLTagList(IN char **XMLP,
				 	  IN int *XMLLengthP,
					  IN char **currentXMLP,
					  tagEntry_t *tagEntryAnchorP)
{
	int retCode;
	tagEntry_t *currentTagEntryP,
		       *nextTagEntryP;
    TRACE_ENTRY("processXMLTagList", ("*XMLP=%.8x, *XMLLengthP=%d, "
		                              "*currentXMLP=%.8x, tagEntryAnchorP=%.8x",
									   *XMLP, *XMLLengthP, *currentXMLP,
									   tagEntryAnchorP));

	currentTagEntryP = tagEntryAnchorP->nextP;

	while(currentTagEntryP)
	{
		TRACE_MID(TRACE_INFO, ("Processing tag element at %.8x", currentTagEntryP));
		retCode = processXMLTagEntry(XMLP,
			                         XMLLengthP,
									 currentXMLP,
									 currentTagEntryP);
		if (retCode != TT_SUCCESS)
		{
		    TRACE_RETURN(RSN_006, retCode);
		}

		/* address next tag name in list */
		currentTagEntryP = currentTagEntryP->nextP;
	}

	/* now free up the tag list */
	currentTagEntryP = tagEntryAnchorP->nextP;
	while(currentTagEntryP)
	{
		TRACE_MID(TRACE_INFO, ("free'ing tag entry at %.8x", currentTagEntryP));
		nextTagEntryP = currentTagEntryP->nextP;
		free(currentTagEntryP);
	    TRACE_FREE("utilities.c-45", currentTagEntryP);
		currentTagEntryP = nextTagEntryP;
	}

    TRACE_RETURN(RSN_007, TT_SUCCESS);
}


/******************************************************************************/
/* convertSyntaxElementsToXML                                                 */
/******************************************************************************/
int convertSyntaxElementsToXML(IN char **XMLP,
						 	   IN int *XMLLengthP,
							   IN char **currentXMLP,
						       IN char *buffer,
						       IN int bufferLength,
						       INOUT char **currentPosP,
							   IN char *endPos,
						       INOUT int *treeDepthP)
{
	elementHeader_t *headerP,
					workHeader,
					*bufHeaderP;

	tagEntry_t tagAnchor = {TAG_ENTRY_DEFAULT},
	           tagAnchorDefault = {TAG_ENTRY_DEFAULT},
			   *anchorEntryP = &tagAnchor,
		       *newTagEntryP,
			   *lastTagEntryP = NULL;
	bool_t elementHasAValue,
		   elementHasAName,
		   elementIsNameValue,
		   keepLooping;
	int dataDepth,
		retCode,
		stringDataLength;
	CciValueType valueType;
	void *dataP,
		 *stringDataP;

    TRACE_ENTRY("convertSyntaxElementsToXML", ("*XMLP=%.8x, "
		                                      "*XMLLengthP=%d, "
		                                      "*currentXMLP=%.8x, "
		                                      "buffer=%.8x, bufferLength=%d, "
		                                      "*currentPosP=%.8x, endPos=%.8x, *treeDepthP=%d",
											  *XMLP, *XMLLengthP,
											  *currentXMLP,
										      buffer, bufferLength, *currentPosP, 
										      endPos, *treeDepthP)); 

	/* keep looping till end of data buffer reached */
	keepLooping = TRUE;
	while(keepLooping)
	{
		/* check we're not at end of serialised data buffer */
		if (*currentPosP == endPos)
		{
			TRACE_MID(TRACE_INFO, ("End of serialised buffer reached"));

			/* write out any closing tags we've stored */
			processXMLTagList(XMLP,
				              XMLLengthP,
							  currentXMLP,
							  anchorEntryP);

			TRACE_RETURN(RSN_062, TT_SUCCESS);
		}

	    bufHeaderP = (elementHeader_t *)(*currentPosP);

	    /* read the next header into local storage to ensure correct alignment */
	    memcpy(&workHeader, *currentPosP, sizeof(elementHeader_t));

	    /* address the next element header */
	    headerP = &workHeader;

		elementHasAValue = FALSE;
		elementHasAName = FALSE;
		elementIsNameValue = FALSE;
		switch (headerP->fixedPart.elementType)
		{
			case CCI_ELEMENT_TYPE_NAME:
				TRACE_MID(TRACE_INFO, ("element type is NAME"));
				elementHasAName = TRUE;
				break;
			case CCI_ELEMENT_TYPE_VALUE:
				TRACE_MID(TRACE_INFO, ("element type is VALUE"));
				elementHasAValue = TRUE;
				break;
			case CCI_ELEMENT_TYPE_NAME_VALUE:
				TRACE_MID(TRACE_INFO, ("element type is NAME_VALUE"));
				elementHasAName = TRUE;
				elementHasAValue = TRUE;
				elementIsNameValue = TRUE;
				break;
			default:
				TRACE_MID(TRACE_ERROR, ("element type is (default) UNKNOWN"));
				TRACE_RETURN(RSN_062, TT_UNKNOWN_ELEMENT_TYPE);
				break;
		}
		
		dataDepth = headerP->fixedPart.depth;

		/* check that we're at the right depth for the serialised data */
		if (dataDepth != *treeDepthP)
		{
			TRACE_MID(TRACE_INFO, ("Current tree level (%d) doesn't match tree level in header (%d)",
									*treeDepthP, dataDepth));

			if (dataDepth < *treeDepthP)
			{
				/* return back to caller */
				TRACE_MID(TRACE_INFO, ("Rewinding tree - return to caller"));

				/* write out any closing tags we've stored */
				processXMLTagList(XMLP,
								  XMLLengthP,
								  currentXMLP,
								  anchorEntryP);
				TRACE_RETURN(RSN_062, TT_SUCCESS);
			}

			/* data depth must be deeper than current depth, so increment depth and recurse */
			TRACE_MID(TRACE_INFO, ("Recursing down one more level"));
			(*treeDepthP)++;
			retCode = convertSyntaxElementsToXML(XMLP,
						 						 XMLLengthP,
												 currentXMLP,
												 buffer,
												 bufferLength,
												 currentPosP,
												 endPos,
												 treeDepthP);
			(*treeDepthP)--;

			/* if we've just processed a VALUE type element, then process last stacked tag */
			if (!elementHasAName && elementHasAValue)
			{
				/* remove last tag entry from list and process it */
				REMOVE_FROM_LINKED_LIST(lastTagEntryP);
				processXMLTagEntry(XMLP,
								   XMLLengthP,
								   currentXMLP,
								   lastTagEntryP);
				free(lastTagEntryP);
			    TRACE_FREE("utilities.c-45", lastTagEntryP);
				lastTagEntryP = NULL;
			}
			continue;
		}

		/* we're at the right level,so extract name and build a tag */
		valueType = headerP->fixedPart.valueType;
		
		if (elementHasAName)
		{
			TRACE_MID(TRACE_INFO, ("element name is '%S'", bufHeaderP->name));
			/* create leading tag for element, either '<xxx>' or '<xxx ' or 'xxx='' */
			if (extendBufferIfRequired(UClen(bufHeaderP->name) + 2,
									   XMLP,
									   currentXMLP,
									   XMLLengthP))
			{
				TRACE_MID(TRACE_INFO, ("Buffer was extended"));
				TRACE_MID(TRACE_INFO, ("*XMLP value is %.8x", *XMLP));
				TRACE_MID(TRACE_INFO, ("*XMLLengthP value is %d", *XMLLengthP));
				TRACE_MID(TRACE_INFO, ("*currentXMLP value is %.8x", *currentXMLP));
			}

			/* create a new tag entry */
			newTagEntryP = (tagEntry_t *)malloc(sizeof(tagEntry_t) +
												UClen(bufHeaderP->name) + 1);
			TRACE_ALLOC("utilities.c-45", newTagEntryP);
			memcpy(newTagEntryP, &tagAnchorDefault, sizeof(tagEntry_t));

			/* address the name buffer - immediately after tag entry structure */
			newTagEntryP->tagName = (char *)((char *)newTagEntryP + sizeof(tagEntry_t));

			/* copy tag name into tag entry */
			UCsprintf(newTagEntryP->tagName, "%S", bufHeaderP->name);

			if (!elementIsNameValue)
			{
				/* add tag to our local tag list */
				INSERT_AFTER_IN_LINKED_LIST(anchorEntryP, newTagEntryP);
				lastTagEntryP = newTagEntryP;
				TRACE_MID(TRACE_INFO, ("Tag added to stack"));
			}

			/* construct the opening brace for leading tag in the XML buffer */
			**currentXMLP = '<';
			(*currentXMLP)++;

			/* copy name into buffer */
			memcpy(*currentXMLP, newTagEntryP->tagName, strlen(newTagEntryP->tagName));
			(*currentXMLP) += strlen(newTagEntryP->tagName);

			/* add closing brace to opening tag in XML buffer */
			**currentXMLP = '>';
			(*currentXMLP)++;
		}

		/* move past the header */
		(*currentPosP) += headerP->fixedPart.headerLength;
		TRACE_MID(TRACE_INFO, ("Buffer pos moved on %d bytes to %.8x",
								headerP->fixedPart.headerLength,
								*currentPosP));

		if (elementHasAValue)
		{
			TRACE_MID(TRACE_INFO, ("Element has a value"));

			/* address the start of the data value in buffer */
			dataP = (void *)(*currentPosP);

			stringDataP = (*currentPosP);
			stringDataLength = strlen((char *)stringDataP);
			TRACE_MID(TRACE_INFO, ("Length of data in buffer is %d", stringDataLength));
			*currentPosP += (stringDataLength+1);  // don't forget the null

			/* copy data into XML buffer */
			if (extendBufferIfRequired(elementIsNameValue ? stringDataLength + 2 :stringDataLength,
									   XMLP,
									   currentXMLP,
									   XMLLengthP))
			{
				TRACE_MID(TRACE_INFO, ("Buffer was extended"));
				TRACE_MID(TRACE_INFO, ("*XMLP value is %.8x", *XMLP));
				TRACE_MID(TRACE_INFO, ("*XMLLengthP value is %d", *XMLLengthP));
				TRACE_MID(TRACE_INFO, ("*currentXMLP value is %.8x", *currentXMLP));
			}

			TRACE_MID(TRACE_INFO, ("Copying %d bytes of data into XML message buffer",
				                    stringDataLength));
			memcpy(*currentXMLP, stringDataP, stringDataLength);
			*currentXMLP += stringDataLength;


			if (elementIsNameValue)
			{
				// add closing tag 
				retCode = processXMLTagEntry(XMLP,
				 	                         XMLLengthP,
					                         currentXMLP,
					                         newTagEntryP);
				if (retCode != TT_SUCCESS)
				{
				    TRACE_RETURN(RSN_061, retCode);
				}

				free(newTagEntryP);
			    TRACE_FREE("utilities.c-45", newTagEntryP);
				lastTagEntryP = NULL;
			}
		}
		else
		{
			TRACE_MID(TRACE_INFO, ("Element has no value"));
		}
	}

    TRACE_RETURN(RSN_062, TT_SUCCESS);
}


/******************************************************************************/
/* getValueFromSyntaxTree                                                     */
/******************************************************************************/
int getValueFromSyntaxTree(IN CciMessage *message,
						   IN CciChar *elementName,
						   IN bool_t lookForChildValueElement,
						   OUT char **valueP,
						   OUT int *valueLengthP)
{
	CciElement *copyFromElement;
	int retCode;

    TRACE_ENTRY("getValueFromSyntaxTree", ("message=%.8x, "
		                                   "elementName=%S, lookForChildValueElement=%s",
										   message, elementName,
										   lookForChildValueElement ? "TRUE" : "FALSE"));
			                     
	/* locate syntax tree element specified by user */
	copyFromElement = locateElementInMessage(message,
											 elementName,
											 FALSE,   /* don't create element */
											 NULL,
											 NULL);
	if (!copyFromElement)
	{
		TRACE_MID(TRACE_INFO, ("Failed to locate specified element in tree"));
		TRACE_RETURN(RSN_001, TT_ELEMENT_NOT_FOUND);
	}

	/* extract the value of this data element */
	retCode = retrieveElementValue(copyFromElement, 
								   lookForChildValueElement,
			                       valueP, 
				                   valueLengthP);
	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_INFO, ("Failed to extract value from specified element in tree"));
		TRACE_RETURN(RSN_001, retCode);
	}

	TRACE_RETURN(RSN_001, TT_SUCCESS);

}


/******************************************************************************/
/* formatDataValueForXML                                                      */
/******************************************************************************/
#define INITIAL_BUFFER_SIZE 1000
int formatDataValueForXML(IN  CciElement *elementP,
					      INOUT char *buffer,
					      IN int bufferLength,
					      OUT int *dataLengthP)
{
	int length,
		rc,
		numBytes,
		localBufferLen = INITIAL_BUFFER_SIZE,
		convBufferLen = INITIAL_BUFFER_SIZE,
		localBuffSize;
	bool_t newBufferAllocated = FALSE,
		   unicodeData = FALSE;
	CciValueType valueType;
	CciBool boolValue;
	CciInt intValue;
	CciReal realValue;
	struct CciTimestamp timeStampValue;
	struct CciTime timeValue;
	struct CciDate dateValue;
    struct CciByteArray byteArrayDetails;
    struct CciBitArray bitArrayDetails;
	CciChar smallBuffer[1];
	char convBuffer[INITIAL_BUFFER_SIZE],
		 localBuffer[INITIAL_BUFFER_SIZE],
		 *convBufferP = convBuffer,
		 *localBufferP = localBuffer;

    TRACE_ENTRY("formatDataValueForXML", ("elementP=%.8x, buffer=%.8x, bufferLength=%d", 
                                           elementP, buffer, bufferLength));

	/* determine the type of the value */
	valueType = cniElementValueType(&rc, elementP);
	TRACE_MID(TRACE_INFO, ("valueType is %d", valueType));

	/* determine how large the buffer needs to be */
	numBytes = 0;
	switch (valueType)
	{
		/****************************************************/ 
		/* The following are the variable length data types */
		/****************************************************/ 
		case CCI_VALUE_TYPE_BYTEARRAY:
			byteArrayDetails.pointer = (void *)smallBuffer;
			byteArrayDetails.size = sizeof(smallBuffer);
			numBytes = cniElementByteArrayValue(&rc,
												elementP,
												&byteArrayDetails);
			// allow for conversion to hexadecimal
			numBytes *= 2;
			break;

		case CCI_VALUE_TYPE_CHAR:
			unicodeData = TRUE;
			numBytes = cniElementCharacterValue(&rc, 
										 		elementP, 
												smallBuffer,
						                        sizeof(smallBuffer));
			// length includes null, so remove it
			numBytes--;
			break;

		case CCI_VALUE_TYPE_BITARRAY:
			bitArrayDetails.pointer = (void *)smallBuffer;
			bitArrayDetails.size = sizeof(smallBuffer);
			numBytes = cniElementBitArrayValue(&rc,
											   elementP,
											   &bitArrayDetails);
			// allow for conversion to hexadecimal
			numBytes *= 2;
			break;

		case CCI_VALUE_TYPE_DECIMAL:
			unicodeData = TRUE;
			numBytes = cniElementDecimalValue(&rc,
									          elementP,
									          smallBuffer,
											  sizeof(smallBuffer));
			// length includes null, so remove it
			numBytes--;
			break;

		/*************************************************/ 
		/* The following are the fixed length data types */
		/*************************************************/ 
		case CCI_VALUE_TYPE_BOOL:
			boolValue = cniElementBooleanValue(&rc, elementP);
			if (boolValue)
			{
				numBytes = strlen("TRUE");
			}
			else
			{
				numBytes = strlen("FALSE");
			}
			break;

		case CCI_VALUE_TYPE_INTEGER:
			intValue = cniElementIntegerValue(&rc, elementP);
			sprintf(localBuffer, "%d", intValue);
			numBytes = strlen(localBuffer);
			break;

		case CCI_VALUE_TYPE_REAL:
			realValue = cniElementRealValue(&rc, elementP);
			sprintf(localBuffer, "%g", realValue);
			numBytes = strlen(localBuffer);
			break;

		case CCI_VALUE_TYPE_TIME:
			timeValue = cniElementTimeValue(&rc, elementP);
			sprintf(localBuffer, "%.2d:%.2d:%g", 
				            timeValue.hour, timeValue.minute, timeValue.second);
			numBytes = strlen(localBuffer);  // hh:mm:ss.sss
			break;

		case CCI_VALUE_TYPE_GMTTIME:
			timeValue = cniElementGmtTimeValue(&rc, elementP);
			sprintf(localBuffer, "%.2d:%.2d:%g", 
				            timeValue.hour, timeValue.minute, timeValue.second);
			numBytes = strlen(localBuffer);  // hh:mm:ss.sss
			break;

		case CCI_VALUE_TYPE_DATE:
			numBytes = 10;  // yyyy-mm-dd
			break;

		case CCI_VALUE_TYPE_TIMESTAMP:
			timeStampValue = cniElementTimestampValue(&rc, elementP);
			sprintf(localBuffer, "%.4d-%.2d-%.2d %.2d:%.2d:%g", 
				             timeStampValue.year, timeStampValue.month, timeStampValue.day,
							 timeStampValue.hour, timeStampValue.minute, timeStampValue.second);
			numBytes = strlen(localBuffer);  // yyyy-mm-dd hh:mm:ss.sss
			break;

		case CCI_VALUE_TYPE_GMTTIMESTAMP:
			timeStampValue = cniElementGmtTimestampValue(&rc, elementP);
			sprintf(localBuffer, "%.4d-%.2d-%.2d %.2d:%.2d:%g", 
				             timeStampValue.year, timeStampValue.month, timeStampValue.day,
							 timeStampValue.hour, timeStampValue.minute, timeStampValue.second);
			numBytes = strlen(localBuffer);  // yyyy-mm-dd hh:mm:ss.sss
			break;

		default:
			TRACE_MID(TRACE_INFO, ("element value type is (default) UNKNOWN"));
		    TRACE_RETURN(RSN_062, TT_ELEMENT_HAS_NO_VALUE);
			break;
	}

	TRACE_MID(TRACE_INFO, ("numBytes set to %d", unicodeData ? numBytes * sizeof(CciChar) :
	                                                           numBytes));

	/* just return the length if no buffer supplied */
	if (!buffer)
	{
		TRACE_MID(TRACE_INFO, ("Returning length %d to caller", numBytes));
		*dataLengthP = numBytes;
	    TRACE_RETURN(RSN_062, TT_SUCCESS);
	}

	/* make sure buffer is large enough to hold the data */
	if (bufferLength < numBytes)
	{
		TRACE_MID(TRACE_ERROR, ("Buffer size (%d) too small to return data (%d)",
			                     bufferLength, numBytes));
		*dataLengthP = numBytes;
	    TRACE_RETURN(RSN_062, TT_BUFFER_TOO_SMALL);
	}

	localBuffSize = unicodeData ? (numBytes+1) * sizeof(CciChar) : numBytes + 1;
	TRACE_MID(TRACE_INFO, ("local buffer size required is %d", localBuffSize));

	/* allocate local buffers, allowing for null character */
	if (localBuffSize > INITIAL_BUFFER_SIZE)
	{
		localBufferP = (char *)malloc(localBuffSize); 
		TRACE_ALLOC("utilities.c-46", localBufferP);
		convBufferP = (char *)malloc(localBuffSize);   

		localBufferLen = localBuffSize;
		convBufferLen = localBuffSize;
		newBufferAllocated = TRUE;
	}

	/* extract element value into buffer */
	switch (valueType)
	{
		case CCI_VALUE_TYPE_BYTEARRAY:
			TRACE_MID(TRACE_INFO, ("element value type is BYTEARRAY"));
			byteArrayDetails.pointer = (void *)localBufferP;
			byteArrayDetails.size = localBufferLen;
			length = cniElementByteArrayValue(&rc,
											  elementP,
											  &byteArrayDetails);

			/* convert into hex chars - dispHex appends a null for us */
			dispHex(localBufferP,
				    length,
					convBufferP);

			/* copy result back to local buffer */
			strcpy(localBufferP, convBufferP);
			TRACE_MID(TRACE_INFO, ("byteArray value is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_CHAR:
			TRACE_MID(TRACE_INFO, ("element value type is CHAR"));
			length = cniElementCharacterValue(&rc, 
										 	  elementP, 
											  (CciChar*)localBufferP,
						                      localBufferLen);
			/* convert into multi-byte string */
            UCToStr((CciChar *)localBufferP,
					convBufferP,
                    convBufferLen,
					NULL);
			/* copy result back to local buffer */
			strcpy(localBufferP, convBufferP);
			TRACE_MID(TRACE_INFO, ("char value is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_BITARRAY:
			TRACE_MID(TRACE_INFO, ("element value type is BITARRAY"));
			bitArrayDetails.pointer = (void *)localBufferP;
			bitArrayDetails.size = localBufferLen;
			length = cniElementBitArrayValue(&rc,
											 elementP,
											 &bitArrayDetails);

			/* convert into hex chars - dispHex appends a null for us */
			dispHex(localBufferP,
				    length,
					convBufferP);

			/* copy result back to local buffer */
			strcpy(localBufferP, convBufferP);
			TRACE_MID(TRACE_INFO, ("bitArray value is '%s'", localBufferP));

			break;

		case CCI_VALUE_TYPE_DECIMAL:
			TRACE_MID(TRACE_INFO, ("element value type is DECIMAL"));
			length = cniElementDecimalValue(&rc,
									        elementP,
									        (CciChar *)localBufferP,
									        localBufferLen);
			*(localBufferP+length) = '\0';
			/* convert into multi-byte string */
            UCToStr((CciChar *)localBufferP,
					convBufferP,
                    convBufferLen,
					NULL);
			/* copy result back to local buffer */
			strcpy(localBufferP, convBufferP);
			TRACE_MID(TRACE_INFO, ("decimal value is '%s'", localBufferP));
			break;

		/*************************************************/ 
		/* The following are the fixed length data types */
		/*************************************************/ 
		case CCI_VALUE_TYPE_BOOL:
			TRACE_MID(TRACE_INFO, ("element value type is BOOL"));
			boolValue = cniElementBooleanValue(&rc, elementP);
			if (boolValue)
			{
				strcpy(localBufferP, "TRUE");
			}
			else
			{
				strcpy(localBufferP, "FALSE");
			}
			TRACE_MID(TRACE_INFO, ("boolValue is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_INTEGER:
			TRACE_MID(TRACE_INFO, ("element value type is INTEGER"));
			intValue = cniElementIntegerValue(&rc, elementP);
			sprintf(localBufferP, "%d", intValue);
			TRACE_MID(TRACE_INFO, ("intValue is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_REAL:
			TRACE_MID(TRACE_INFO, ("element value type is REAL"));
			realValue = cniElementRealValue(&rc, elementP);
			sprintf(localBufferP, "%g", realValue);
			TRACE_MID(TRACE_INFO, ("realValue is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_TIME:
			TRACE_MID(TRACE_INFO, ("element value type is TIME"));
			timeValue = cniElementTimeValue(&rc, elementP);
			sprintf(localBufferP, "%.2d:%.2d:%g", 
				            timeValue.hour, timeValue.minute, timeValue.second);
			TRACE_MID(TRACE_INFO, ("timeValue is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_GMTTIME:
			TRACE_MID(TRACE_INFO, ("element value type is GMTTIME"));
			timeValue = cniElementGmtTimeValue(&rc, elementP);
			sprintf(localBufferP, "%.2d:%.2d:%g", 
				            timeValue.hour, timeValue.minute, timeValue.second);
			TRACE_MID(TRACE_INFO, ("timeValue is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_DATE:
			TRACE_MID(TRACE_INFO, ("element value type is DATE"));
			dateValue = cniElementDateValue(&rc, elementP);
			sprintf(localBufferP, "%.4d-%.2d-%.2d", 
				            dateValue.year, dateValue.month, dateValue.day);
			TRACE_MID(TRACE_INFO, ("dateValue is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_TIMESTAMP:
			TRACE_MID(TRACE_INFO, ("element value type is TIMESTAMP"));
			timeStampValue = cniElementTimestampValue(&rc, elementP);
			sprintf(localBufferP, "%.4d-%.2d-%.2d %.2d:%.2d:%g", 
				             timeStampValue.year, timeStampValue.month, timeStampValue.day,
							 timeStampValue.hour, timeStampValue.minute, timeStampValue.second);
			TRACE_MID(TRACE_INFO, ("timeStampValue is '%s'", localBufferP));
			break;

		case CCI_VALUE_TYPE_GMTTIMESTAMP:
			TRACE_MID(TRACE_INFO, ("element value type is GMTTIMESTAMP"));
			timeStampValue = cniElementGmtTimestampValue(&rc, elementP);
			sprintf(localBufferP, "%.4d-%.2d-%.2d %.2d:%.2d:%g", 
				             timeStampValue.year, timeStampValue.month, timeStampValue.day,
							 timeStampValue.hour, timeStampValue.minute, timeStampValue.second);
			TRACE_MID(TRACE_INFO, ("timeStampValue is '%s'", localBufferP));
			break;

		default:
			TRACE_MID(TRACE_INFO, ("element value type is (default) UNKNOWN"));
		    TRACE_RETURN(RSN_062, TT_ELEMENT_HAS_NO_VALUE);
			break;
	}

	/* copy extracted data into users buffer */
	memcpy(buffer, localBufferP, numBytes);

	if (newBufferAllocated)
	{
		free(localBufferP);
	    TRACE_FREE("utilities.c-46", localBufferP);
		free(convBufferP);
	    TRACE_FREE("utilities.c-47", convBufferP);
	}

	TRACE_MID(TRACE_INFO, ("Returning length %d to caller", numBytes));
	*dataLengthP = numBytes;

	TRACE_RETURN(RSN_062, TT_SUCCESS);
}


/******************************************************************************/
/* messageBodyIsXML                                                           */
/******************************************************************************/
bool_t messageBodyIsXML(IN CciMessage *message)
{
	CciElement *rootElementP,
		       *childElementP;
	CciElementType elementType;
	int rc;
    TRACE_ENTRY("messageBodyIsXML", ("message=%.8x", message));

	/* 
	 * Address root element, then last child
	 */
    rootElementP = cniRootElement(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("cniRootElement for message FAILED"));
		TRACE_RETURN_TYPE(RSN_069, bool_t, FALSE);
	}
	TRACE_MID(TRACE_INFO, ("Addressed  message root element at %.8x", 
		                    rootElementP));

	/* address last child element */
	childElementP = cniLastChild(&rc,
                                 rootElementP);
	if (!childElementP)
	{
		/* body not found in message - generate error */
		TRACE_MID(TRACE_ERROR, ("Failed to address last child "
								"(cniLastChild rc=%d)", rc));
		TRACE_RETURN_TYPE(RSN_069, bool_t, FALSE);
	}

	/* get the element type */
	elementType = cniElementType(&rc, 
		                         childElementP);
    if (elementType == MJB_ELEMENT_TYPE_XML_BODY)
	{
		TRACE_MID(TRACE_INFO, ("Message body is XML format"));
		TRACE_RETURN_TYPE(RSN_069, bool_t, TRUE);
	}

	TRACE_RETURN_TYPE(RSN_069, bool_t, FALSE);
}


/******************************************************************************/
/* setMessageFormat                                                           */
/******************************************************************************/
int setMessageFormat(IN CciMessage *message,
	                 IN CciChar *format)
{
	int rc;
	CciElement *mqmdElementP,
	           *formatElementP;

    TRACE_ENTRY("setMessageFormat", ("message=%.8x, format=%S",
		                              message, format));
	
	/* locate MQMD element in message... */
	mqmdElementP = getMessageElement(message, MQMD_ELEMENT);
	if (!mqmdElementP)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to address MQMD element"));
		TRACE_RETURN(RSN_001, TT_FAILED_TO_GET_MESSAGE_MQMD);
	}
	TRACE_MID(TRACE_INFO, ("Addressed MQMD element at %.8x", mqmdElementP));

	/* locate Format element in MQMD... */
	formatElementP = cniSearchFirstChild(&rc,
		                                 mqmdElementP,
							             CCI_COMPARE_MODE_NAME,
							             CCI_ELEMENT_TYPE_NAME_VALUE,
										 U.FORMAT);

    if (!formatElementP)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to address Format element"));
		TRACE_RETURN(RSN_002, TT_FAILED_TO_GET_MESSAGE_FORMAT);
	}
	TRACE_MID(TRACE_INFO, ("Addressed Format element at %.8x", formatElementP));

	cniSetElementCharacterValue(&rc,
		                        formatElementP,
								format,
								MQ_FORMAT_LENGTH);
	if (rc != CCI_SUCCESS)
	{
	    TRACE_MID(TRACE_ERROR, ("Failed to set Format "
			                    "(cniSetElementByteArrayValue rc=%d)", rc));
	    TRACE_RETURN(RSN_055, TT_FAILED_TO_SET_FORMAT_IN_MESSAGE);
	}

    TRACE_RETURN(RSN_001, TT_SUCCESS);
}


#if 0
void testAddress(int label)
{
	int i;
	aggregateMember_t *memberP;

    TRACE_ENTRY("testAddress", ("label=%d", label));

	if (!addrListP)
	{
		TRACE_RETURN_VOID(RSN_001);
	}

	/* run through each of the addresses in the list and check each one */
	for (i=0; i<addrListP->nextAddress; i++)
	{
		memberP = (aggregateMember_t *)(addrListP->address[i]);
		TRACE_MID(TRACE_INFO, ("memberP (%.8x) contains %.8x",memberP, memberP->messageP));
	}

	TRACE_RETURN_VOID(RSN_001);
}
#endif

